from pygarl.middlewares import PlotterMiddleware
from pygarl.mocks import VerboseMiddleware
from pygarl.data_readers import SerialDataReader, FileDataReader
from pygarl.sample_managers import DiscreteSampleManager

# This example uses a FileDataReader to read data from a stream file
# and plot the recorded sample


def run_example(*args, **kwargs):
    # Create the SerialDataReader
    fdr = FileDataReader(args[0], verbose=False)

    # Create the SampleManager
    manager = DiscreteSampleManager()

    # Attach the manager
    fdr.attach_manager(manager)

    # Create the VerboseMiddleware that prints the received sample
    middleware = PlotterMiddleware(blocking=True)

    # Attach the middleware
    manager.attach_receiver(middleware)

    # Open the serial connection
    fdr.open()
    print("Opened!")

    # Start the main loop
    fdr.mainloop()


if __name__ == '__main__':
    run_example("D:/sample.txt")



import unittest
import scipy as sp
from pygarl.abstracts import *
from pygarl.predictors import *
from pygarl.mocks import *
from pygarl.base import *

# To execute tests, go to the project main directory and type:
# python -m unittest discover


class HighestAxisPredictorTestCase(unittest.TestCase):
    """
    Tests to check HighestAxisPredictor behaviour
    """
    def setUp(self):
        # Initialize the HighestAxisPredictor
        self.predictor = HighestAxisPredictor()

    def tearDown(self):
        # Destroy the predictor
        self.predictor = None

    def test_predict_only_one_sample_frame(self):
        sample = Sample(data=[[1, 2, 3, 2]])

        self.assertEqual(self.predictor.predict(sample), "2")

    def test_predict_only_one_sample_frame_with_negatives(self):
        sample = Sample(data=[[1, -2, -3, -2]])

        self.assertEqual(self.predictor.predict(sample), "0")

    def test_predict_absolute_values_option(self):
        sample = Sample(data=[[1, -2, -3, -2]])

        # Set the absolute values option
        self.predictor.absolute_values = True

        self.assertEqual(self.predictor.predict(sample), "2")

    def test_predict_multiple_frames(self):
        sample = Sample(data=[[2, 3, 3, 2], [0, 0, 3, 1]])

        self.assertEqual(self.predictor.predict(sample), "2")

    def test_predict_empty_sample_should_raise_error(self):
        sample = Sample(data=[[]])

        self.assertRaises(ValueError, self.predictor.predict, sample)

if __name__ == '__main__':
    unittest.main()



import serial


# Plot the input received from the specified serial port
def sprint(port, baudrate):
    print("Opening serial port...")
    s = serial.Serial(port=port, baudrate=baudrate, timeout=1)

    # Start the endless loop
    while True:
        # Read a line from the serial connection
        line = s.readline()

        # Replace the ending characters
        line = line.replace("\n", "")

        # Print the line
        print(line)



from __future__ import print_function
from pygarl.classifiers import SVMClassifier, MLPClassifier
import sys


def train_classifier(classifier, dataset_dir, output_file, n_jobs=1, **kwargs):
    """
    Train a model using the passed classifier from the given dataset and save it to a file.
    
    :param classifier: Classifier used to create a model
    :param dataset_dir: Path of the dataset directory containing the samples
    :param output_file: Output file of the model 
    """
    # Load the data
    print("Loading the data...", end="")

    classifier.load()
    print("LOADED!")

    print("Training the model...")

    # Train the model and obtain the score
    score = classifier.train_model()

    print("FINAL SCORE:", score)

    print("Saving the model to the output file:", output_file)

    classifier.save_model(output_file)

    print("DONE")

    # Plot the confusion matrix
    classifier.plot_confusion_matrix()


def train_svm_classifier(dataset_dir, output_file, n_jobs=1):
    """
    Train an SVM model from the given dataset and save it to a file.
    """
    # Create the classifier
    classifier = SVMClassifier(dataset_path=dataset_dir, verbose=True, n_jobs=n_jobs,
                               autoscale_size=50)

    # Train the classifier
    train_classifier(classifier=classifier, dataset_dir=dataset_dir, output_file=output_file, n_jobs=n_jobs)


def train_mlp_classifier(dataset_dir, output_file, n_jobs=1):
    """
    Train an MLP model from the given dataset and save it to a file.
    """
    # Create the classifier
    classifier = MLPClassifier(dataset_path=dataset_dir, verbose=True, n_jobs=n_jobs,
                               autonormalize=True, autoscale_size=15)

    # Train the classifier
    train_classifier(classifier=classifier, dataset_dir=dataset_dir, output_file=output_file, n_jobs=n_jobs)


# If launched directly, parse the parameters from sys
if __name__ == '__main__':
    train_svm_classifier(sys.argv[0], sys.argv[1], 8)



import unittest
import scipy as sp

from pygarl.sample_managers import DiscreteSampleManager, StreamSampleManager
from pygarl.abstracts import *
from pygarl.mocks import *
from pygarl.base import *

# To execute tests, go to the project main directory and type:
# python -m unittest discover


class DiscreteSampleManagerTestCase(unittest.TestCase):
    """
    Tests to check DiscreteSampleManager behaviour
    """
    def setUp(self):
        # Initialize the DiscreteSampleManager
        self.manager = DiscreteSampleManager(min_sample_length=1)

    def tearDown(self):
        # Destroy the sample manager
        self.manager = None

    def test_sample_received_correctly(self):
        # Initialize a Mock Receiver
        receiver = MockReceiver()
        self.assertIsNone(receiver.received_sample)

        # Attach the receiver
        self.manager.attach_receiver(receiver)

        # No data at the beginning
        self.assertEqual(len(self.manager.buffer), 0)
        # Send the START signal
        self.manager.receive_signal(ControlSignal.START)
        # Still no data
        self.assertEqual(len(self.manager.buffer), 0)

        # Send some data
        self.manager.receive_data([1, 2, 3])
        self.manager.receive_data([4, 5, 6])
        self.manager.receive_data([7, 8, 9])

        # Check the data has been received
        self.assertEqual(len(self.manager.buffer), 3)

        # Send the STOP signal
        self.manager.receive_signal(ControlSignal.STOP)

        # Check if the Sample has been received correctly by the receiver
        self.assertIsNotNone(receiver.received_sample)
        # Check if the arrays are the same
        self.assertTrue(sp.allclose(receiver.received_sample.data, sp.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])))

    def test_sample_with_min_sample_length_should_fail(self):
        # Initialize a Mock Receiver
        receiver = MockReceiver()
        self.assertIsNone(receiver.received_sample)

        self.manager.min_sample_length = 10  # Set the minimum number of frames to 10

        # Attach the receiver
        self.manager.attach_receiver(receiver)

        # No data at the beginning
        self.assertEqual(len(self.manager.buffer), 0)
        # Send the START signal
        self.manager.receive_signal(ControlSignal.START)
        # Still no data
        self.assertEqual(len(self.manager.buffer), 0)

        # Send some data
        self.manager.receive_data([1, 2, 3])
        self.manager.receive_data([4, 5, 6])
        self.manager.receive_data([7, 8, 9])

        # Check the data has been received
        self.assertEqual(len(self.manager.buffer), 3)

        # Send the STOP signal
        self.manager.receive_signal(ControlSignal.STOP)

        # The sample doesn't have enough frames, so the receiver should not be notified
        self.assertIsNone(receiver.received_sample)

    def test_start_again_should_reset_buffer(self):
        # Initialize a Mock Receiver
        receiver = MockReceiver()
        self.assertIsNone(receiver.received_sample)

        # Attach the receiver
        self.manager.attach_receiver(receiver)

        # No data at the beginning
        self.assertEqual(len(self.manager.buffer), 0)
        # Send the START signal
        self.manager.receive_signal(ControlSignal.START)
        # Still no data
        self.assertEqual(len(self.manager.buffer), 0)

        # Send some data
        self.manager.receive_data([1, 2, 3])
        self.manager.receive_data([4, 5, 6])
        self.manager.receive_data([7, 8, 9])

        # Check the data has been received
        self.assertEqual(len(self.manager.buffer), 3)

        # Send the STOP signal
        self.manager.receive_signal(ControlSignal.STOP)

        # Check if the Sample has been received correctly by the receiver
        self.assertIsNotNone(receiver.received_sample)
        # Check if the arrays are the same
        self.assertTrue(sp.allclose(receiver.received_sample.data, sp.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])))

        # Start a new sample
        self.manager.receive_signal(ControlSignal.START)
        # No data
        self.assertEqual(len(self.manager.buffer), 0)


class StreamSampleManagerTestCase(unittest.TestCase):
    """
    Tests to check StreamSampleManager behaviour
    """
    def setUp(self):
        # Initialize the StreamSampleManager
        self.manager = StreamSampleManager(window=4, step=2)

    def tearDown(self):
        # Destroy the sample manager
        self.manager = None

    def test_sample_received_when_window_size_is_reached(self):
        # Initialize a Mock Receiver
        receiver = MockReceiver()
        self.assertIsNone(receiver.received_sample)

        # Attach the receiver
        self.manager.attach_receiver(receiver)

        # No data at the beginning
        self.assertEqual(len(self.manager.buffer), 0)
        # Send the START signal
        self.manager.receive_signal(ControlSignal.START)
        # Still no data
        self.assertEqual(len(self.manager.buffer), 0)

        # Send some data
        self.manager.receive_data([1, 2, 3])
        self.manager.receive_data([4, 5, 6])
        self.manager.receive_data([7, 8, 9])
        self.manager.receive_data([10, 11, 12])

        # Check if the Sample has been received correctly by the receiver
        self.assertIsNotNone(receiver.received_sample)
        # Check if the arrays are the same
        self.assertTrue(sp.allclose(receiver.received_sample.data,
                                    sp.array([[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12]])))

    def test_step_working_correctly(self):
        # Initialize a Mock Receiver
        receiver = MockReceiver()
        self.assertIsNone(receiver.received_sample)

        # Attach the receiver
        self.manager.attach_receiver(receiver)

        # No data at the beginning
        self.assertEqual(len(self.manager.buffer), 0)
        # Send the START signal
        self.manager.receive_signal(ControlSignal.START)
        # Still no data
        self.assertEqual(len(self.manager.buffer), 0)

        # Send some data
        self.manager.receive_data([1, 2, 3])
        self.manager.receive_data([4, 5, 6])
        self.manager.receive_data([7, 8, 9])
        self.manager.receive_data([10, 11, 12])

        # Check if the Sample has been received correctly by the receiver
        self.assertIsNotNone(receiver.received_sample)
        # Check if the arrays are the same
        self.assertTrue(sp.allclose(receiver.received_sample.data,
                                    sp.array([[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12]])))

        self.manager.receive_data([13, 14, 15])
        self.manager.receive_data([16, 17, 18])

        # Check the new sample data
        self.assertTrue(sp.allclose(receiver.received_sample.data,
                                    sp.array([[7, 8, 9], [10, 11, 12], [13, 14, 15], [16, 17, 18]])))

    def test_sample_stop_before_reaching_window_size(self):
        # Initialize a Mock Receiver
        receiver = MockReceiver()
        self.assertIsNone(receiver.received_sample)

        # Attach the receiver
        self.manager.attach_receiver(receiver)

        # No data at the beginning
        self.assertEqual(len(self.manager.buffer), 0)
        # Send the START signal
        self.manager.receive_signal(ControlSignal.START)
        # Still no data
        self.assertEqual(len(self.manager.buffer), 0)

        # Send some data
        self.manager.receive_data([1, 2, 3])
        self.manager.receive_data([4, 5, 6])
        self.manager.receive_data([7, 8, 9])

        # Check the data has been received
        self.assertEqual(len(self.manager.buffer), 3)

        # Send the STOP signal
        self.manager.receive_signal(ControlSignal.STOP)

        # Check if the Sample has been received correctly by the receiver
        self.assertIsNotNone(receiver.received_sample)
        # Check if the arrays are the same
        self.assertTrue(sp.allclose(receiver.received_sample.data, sp.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])))

    def test_step_must_be_lower_or_equal_to_window_size(self):
        # Check the StreamSampleManager raises an exception if step > window
        self.assertRaises(ValueError, StreamSampleManager, step=10, window=5)

if __name__ == '__main__':
    unittest.main()



import unittest
import os
import shutil
from pygarl.abstracts import *
from pygarl.mocks import *
from pygarl.base import *

# To execute tests, go to the project main directory and type:
# python -m unittest discover


class SampleTestCase(unittest.TestCase):
    """
    Tests to check FileGestureRecorder behaviour
    """

    def setUp(self):
        # Create a test directory if it doesn't exists
        if not os.path.exists("test_sample_dir"):
            os.makedirs("test_sample_dir")

        # Create the sample
        self.sample = Sample([[1, 2, 3], [4, 5, 6]], gesture_id="TESTSAMPLE")

    def tearDown(self):
        # Destroy the test directory
        shutil.rmtree("test_sample_dir")

        # Destroy the sample
        self.recorder = None

    def test_data_must_be_a_2_dimensional_array(self):
        self.assertRaises(ValueError, Sample, [])

    def test_save_to_file(self):
        filepath = os.path.join("test_sample_dir", "test_sample.txt")
        # Save the sample to file
        self.sample.save_to_file(filepath)

        # Open the json file
        with open(filepath) as data_file:
            data = json.load(data_file)

        self.assertEqual(data['gesture_id'], self.sample.gesture_id)
        self.assertEqual(data['data'], self.sample.data.tolist())

        # Remove the sample at the end
        os.remove(filepath)

    def test_load_from_file(self):
        # Create the sample file...

        filepath = os.path.join("test_sample_dir", "test_sample.txt")
        # Save the sample to file
        self.sample.save_to_file(filepath)

        # Load the sample
        sample = Sample.load_from_file(filepath)

        self.assertEqual(sample.gesture_id, self.sample.gesture_id)
        self.assertEqual(sample.data.tolist(), self.sample.data.tolist())

        # Remove the sample at the end
        os.remove(filepath)

    def test_scale_frames(self):
        sample = Sample(data=[[0, 0, 0], [1, 2, 4], [2, 4, 8], [3, 6, 12], [4, 8, 16]])
        self.assertEqual(sample.data.tolist(), [[0, 0, 0], [1, 2, 4], [2, 4, 8], [3, 6, 12], [4, 8, 16]])

        # Scale down the frames
        sample.scale_frames(3)

        self.assertEqual(sample.data.tolist(), [[0, 0, 0], [2, 4, 8], [4, 8, 16]])

    def test_get_linearized(self):
        sample = Sample(data=[[0, 0, 0], [1, 2, 4], [2, 4, 8]])
        self.assertEqual(sample.data.tolist(), [[0, 0, 0], [1, 2, 4], [2, 4, 8]])

        self.assertEqual(sample.get_linearized().tolist(), [[0, 0, 0, 1, 2, 4, 2, 4, 8]])
        self.assertEqual(sample.get_linearized(one_dimensional=True).tolist(), [0, 0, 0, 1, 2, 4, 2, 4, 8])


class CallbackManagerTestCase(unittest.TestCase):
    """
    Tests to check CallbackManager correct behaviour
    """
    def setUp(self):
        # Initialize the callback manager
        self.manager = CallbackManager()

    def tearDown(self):
        # Destroy the callback manager
        self.manager = None

    def test_callback_attached_correctly(self):
        receiver = MockFunctionReceiver()

        # Should not be attached initially
        self.assertFalse('my_gesture' in self.manager.callbacks)
        # Attach the function
        self.manager.attach_callback('my_gesture', receiver.receive)
        # Should be attached
        self.assertTrue('my_gesture' in self.manager.callbacks)

    def test_callback_detached_correctly(self):
        receiver = MockFunctionReceiver()

        # Attach the callback function
        self.manager.attach_callback('my_gesture', receiver.receive)
        # Should be attached
        self.assertTrue('my_gesture' in self.manager.callbacks)
        # Detach the callback function
        self.manager.detach_callback('my_gesture')
        # Should be detached
        self.assertFalse('my_gesture' in self.manager.callbacks)

    def test_callback_notify_gesture(self):
        receiver = MockFunctionReceiver()

        # Attach the callback function
        self.manager.attach_callback('my_gesture', receiver.receive)
        # Initially it's not received
        self.assertFalse(receiver.received)
        # Send the notify
        self.manager.notify_gesture('my_gesture')
        # Should be received
        self.assertTrue(receiver.received)

    def test_callback_notify_gesture_should_not_call_the_function(self):
        receiver = MockFunctionReceiver()

        self.manager.attach_callback('my_gesture', receiver.receive)
        self.assertFalse(receiver.received)
        # Notify another gesture
        self.manager.notify_gesture('not_my_gesture')
        # Should not be called
        self.assertFalse(receiver.received)

    def test_callback_receive_gesture_forward_the_event_to_the_callbacks(self):
        receiver = MockFunctionReceiver()

        self.manager.attach_callback('my_gesture', receiver.receive)
        self.assertFalse(receiver.received)
        # Simulate an event, should propagate to the attached callbacks
        self.manager.receive_gesture('my_gesture')
        self.assertTrue(receiver.received)

    def test_callback_notify_gesture_multiple_callbacks(self):
        """
        Test to verify that multiple gestures call the corresponding callbacks
        """
        receiver1 = MockFunctionReceiver()
        receiver2 = MockFunctionReceiver()

        # Attach the gestures
        self.manager.attach_callback('my_gesture', receiver1.receive)
        self.manager.attach_callback('my_gesture2', receiver2.receive)
        self.assertFalse(receiver1.received)
        self.assertFalse(receiver2.received)

        # Only the first one has been called
        self.manager.notify_gesture('my_gesture')
        self.assertTrue(receiver1.received)
        self.assertFalse(receiver2.received)

        # Both has been called
        self.manager.notify_gesture('my_gesture2')
        self.assertTrue(receiver1.received)
        self.assertTrue(receiver2.received)

if __name__ == '__main__':
    unittest.main()



from __future__ import print_function
from pygarl.base import Sample
from pygarl.abstracts import AbstractMiddleware
import numpy as np


class GradientThresholdMiddleware(AbstractMiddleware):
    """
    Used to extract gestures from a stream of samples.
    It calculates the gradient of the sample data and the average value,
    if this value is superior to the fixed threshold, it make the sample pass,
    if not, it blocks the sample.
    If the group=True, the middleware will try to group different samples into one.
    """
    def __init__(self, threshold=10, group=False, sample_group_delay=2, verbose=False, autotrim=False, trim_threshold=10):
        """
        Class constructor
        :param threshold: The value that must be crossed to mark a sample as valid.
                          The higher the value, the most intense the movement must be to
                          pass the middleware.
                          
        :param group:     If True, the middleware tries to group samples belonging to the same movement
                          into one sample.
                          
        :param: sample_group_delay: The amount of samples that are included in the grouped sample that
                                    have not crossed the threshold. This is useful because sometimes
                                    samples have some oscillations that make the middleware split
                                    samples even if they belong to the same one.
        
        :param verbose:   If True, prints more information.
        
        :param autotrim: If True, trim the sample data ends if they are below the trim_threshold.
                         Note: works only if group = True
        
        :param trim_threshold: Threshold for the autotrim property.
        """
        # Call the base constructor
        AbstractMiddleware.__init__(self)

        # Set the parameters
        self.threshold = threshold
        self.verbose = verbose
        self.group = group
        self.sample_group_delay = sample_group_delay
        self.autotrim = autotrim
        self.trim_threshold = trim_threshold

        # Create the array that will hold the sample data to be grouped
        # Initially it is None
        self.buffer = None

        # Counter for the sample_group_delay property
        self.sample_group_delay_counter = 0

    def add_sample_to_buffer(self, sample):
        """
        Add the given sample to the buffer to group them.
        :param sample: Sample to add.
        """
        # Get a copy of the sample data
        sample_data = np.array(sample.data, copy=True)

        # If buffer is empty, it becomes equal to the sample data
        if self.buffer is None:
            self.buffer = sample_data
        else:  # buffer is not empty, concatenate the new sample data
            self.buffer = np.concatenate((self.buffer, sample_data))

    def delete_buffer(self):
        """
        Delete the current buffer by setting it to None
        """
        self.buffer = None

    def process_sample(self, sample):
        # Get the sample gradient
        gradient = sample.gradient()

        # Calculate the average
        average = np.average(gradient)

        # Calculate the absolute value
        average = np.absolute(average)

        # If verbose, print the average
        if self.verbose:
            print("GTM Average:", average)

        # If true, the sample will pass.
        is_allowed_to_pass = False

        # If the average is bigger than the threshold, return the sample
        # If not, suppress the sample by returning None
        if average >= self.threshold:
            is_allowed_to_pass = True

            # Reset the sample_group_delay counter
            self.sample_group_delay_counter = 0
        else:
            if self.group and self.sample_group_delay_counter < self.sample_group_delay:
                is_allowed_to_pass = True

                # Increase the sample_group_delay counter
                self.sample_group_delay_counter += 1

        if is_allowed_to_pass:
            # If grouping is not enabled, return the sample directly
            if not self.group:
                return sample
            else:  # grouping is enabled, return None and add the current sample to the buffer
                self.add_sample_to_buffer(sample)

                # If verbose, print some info
                if self.verbose:
                    print("GTM:", "Suppressed sample to be grouped.")

                # Suppress the single sample
                return None
        else:
            # If grouping is not enabled, suppress the sample
            if not self.group:
                # Suppress the Sample
                return None
            else:  # Grouping is enabled
                # Check that the buffer is not empty
                if self.buffer is not None:
                    # If this sample has not crossed the threshold it means that the grouped sample
                    # is finished and can be returned.

                    # Copy the buffer
                    grouped_data = np.copy(self.buffer)

                    # Delete the buffer
                    self.delete_buffer()

                    # Create a new sample with the grouped data
                    new_sample = Sample(data=grouped_data, gesture_id=sample.gesture_id)

                    # Trim the sample data if autotrim is enabled
                    if self.autotrim:
                        new_sample.trim(self.trim_threshold)

                    # Return the new sample
                    return new_sample
                else:  # Buffer empty, return none
                    return None


class AbsoluteScaleMiddleware(AbstractMiddleware):
    """
    It scales and calculates the absolute values of the sample data.
    """
    def __init__(self, scale_size=50, subtract=None, rolling_mean_window=None):
        # Call the base constructor
        AbstractMiddleware.__init__(self)

        # Set the parameters
        self.scale_size = scale_size
        self.subtract = subtract
        self.rolling_mean_window = rolling_mean_window

    def process_sample(self, sample):
        """
        Scale and calculate the sample data
        """
        # Subtract if enabled
        if self.subtract is not None:
            sample.subtract(self.subtract)

        # Calculate the absolute value
        sample.abs()

        # Calculate the rolling mean if enabled
        if self.rolling_mean_window is not None:
            sample.rolling_mean(self.rolling_mean_window)

        # Scale the data
        sample.scale_frames(n_frames=self.scale_size)

        return sample


class FFTMiddleware(AbstractMiddleware):
    """
    Replace the sample with his fourier transform
    """
    def __init__(self):
        # Call the base constructor
        AbstractMiddleware.__init__(self)

    def process_sample(self, sample):
        """
        Calculates the fft and replace the sample data
        """
        # Calculate the FFT and replace the data
        sample.fft()

        return sample


class TrimmerMiddleware(AbstractMiddleware):
    """
    Trim the extremes of a sample that are lower than the threshold
    """
    def __init__(self, threshold=300):
        # Call the base constructor
        AbstractMiddleware.__init__(self)

        # Set the parameters
        self.threshold = threshold

    def process_sample(self, sample):
        """
        Trim the sample data
        """
        # Trim the sample data
        sample.trim(self.threshold)

        return sample


class LengthThresholdMiddleware(AbstractMiddleware):
    """
    Let pass the samples that are between min_len and max_len, block the others.
    """
    def __init__(self, min_len=0, max_len=100000, verbose=False):
        # Call the base constructor
        AbstractMiddleware.__init__(self)

        # Set the parameters
        self.min_len = min_len
        self.max_len = max_len
        self.verbose = verbose

    def process_sample(self, sample):
        """
        Filter the sample based on the length
        """
        # Print the sample length
        if self.verbose:
            print("SAMPLE_LEN", sample.framelen())

        # If the sample is between, let it pass
        # block it otherwise
        if self.min_len <= sample.framelen() <= self.max_len:
            return sample
        else:
            if self.verbose:
                print("LTM: Blocked sample.")
            return None

class DelayGrouperMiddleware(AbstractMiddleware):
    """
    Group received samples into one Sample if they arrive within "delay" time from one another.
    Useful to group individual samples spawned by the GradientThresholdMiddleware that belongs
    to the same gesture.
    """
    def __init__(self, delay=300):
        """
        Class constructor
        :param delay: Milliseconds between samples that must be grouped.
        """
        # Call the base constructor
        AbstractMiddleware.__init__(self)

        # Set the parameters
        self.delay = delay

    def process_sample(self, sample):
        # Not implemented yet
        raise NotImplementedError("DelayGrouperMiddleware is not implemented yet")


class PlotterMiddleware(AbstractMiddleware):
    """
    Used to plot sample after being received
    """
    def __init__(self, blocking=False):
        # Call the base constructor
        AbstractMiddleware.__init__(self)

        self.blocking = blocking

    def process_sample(self, sample):
        """
        Plot the sample and return it
        :param sample: sample to plot
        """
        # Plot the sample
        sample.plot(self.blocking)

        return sample


from pygarl.classifiers import MLPClassifier, SVMClassifier
from pygarl.middlewares import AbsoluteScaleMiddleware
from pygarl.plugins.train import train_classifier


def train(dataset_dir, output_file, n_jobs=1):
    """
    Used to train the knock sensor data.
    Train an MLP model from the given dataset and save it to a file.
    """
    asm = AbsoluteScaleMiddleware()

    # Create the classifier
    classifier = SVMClassifier(dataset_path=dataset_dir, verbose=True, n_jobs=n_jobs,
                               autonormalize=True, autoscale_size=40, middlewares=[asm])

    # Train the classifier
    train_classifier(classifier=classifier, dataset_dir=dataset_dir, output_file=output_file, n_jobs=n_jobs)



from pygarl.abstracts import AbstractMiddleware
from pygarl.base import CallbackManager
from pygarl.mocks import VerboseMiddleware
from pygarl.predictors import HighestAxisPredictor
from pygarl.data_readers import SerialDataReader
from pygarl.sample_managers import DiscreteSampleManager, StreamSampleManager

# This example uses a SerialDataReader to read data from a serial port
# and uses a HighestAxisPredictor to print the most relevant axis


def run_example(*args, **kwargs):
    # Create the SerialDataReader
    sdr = SerialDataReader(kwargs['port'], expected_axis=6, verbose=False)

    # Create the SampleManager
    manager = StreamSampleManager(window=10, step=5)

    # Attach the manager
    sdr.attach_manager(manager)

    # Create the VerboseMiddleware that prints the received sample
    middleware = VerboseMiddleware(verbose=False)

    # Attach the middleware
    manager.attach_receiver(middleware)

    # Create a predictor
    predictor = HighestAxisPredictor(absolute_values=True)

    # Attach the predictor
    middleware.attach_receiver(predictor)

    # Create a callback manager
    callback_mg = CallbackManager(verbose=True)

    # Attach the callback manager
    predictor.attach_callback_manager(callback_mg)

    # Open the serial connection
    sdr.open()
    print("Opened!")

    # Start the main loop
    sdr.mainloop()



from __future__ import print_function
import os, random
from pygarl.data_readers import SerialDataReader
from pygarl.middlewares import GradientThresholdMiddleware, PlotterMiddleware, LengthThresholdMiddleware
from pygarl.recorders import FileGestureRecorder
from pygarl.sample_managers import DiscreteSampleManager, StreamSampleManager

# python -m pygarl record -p COM6 -d D:\GitHub\pygarl-datasets\finger_dataset -g right -m stream
from pygarl.utils import RandomGestureChooser


def record_new_samples(port, gesture_id, target_dir, expected_axis):
    """
    Used to record new samples for the specified gesture_id
    """
    print("RECORDING NEW SAMPLES")

    # Create the target directory if it doesn't exists
    if not os.path.exists(target_dir):
        os.makedirs(target_dir)

    print("SERIAL PORT:", port)
    print("TARGET DIRECTORY:", target_dir)
    print("GESTURE ID:", gesture_id)

    # Create the SerialDataReader
    sdr = SerialDataReader(port, expected_axis=expected_axis, verbose=False)

    # Create the SampleManager
    manager = DiscreteSampleManager()

    # Attach the manager
    sdr.attach_manager(manager)

    # Create a FileGestureRecorder
    recorder = FileGestureRecorder(target_dir=target_dir, forced_gesture_id=gesture_id, verbose=True)

    # Attach the recorder to the manager
    manager.attach_receiver(recorder)

    # Open the serial connection
    print("Opening serial port...")
    sdr.open()
    print("Opened!")

    print("To exit the loop, press Ctrl+C.")

    # Start the main loop
    sdr.mainloop()


def record_new_samples_stream(port, gesture_id, target_dir, expected_axis, threshold=50):
    """
    Used to record new samples for the specified gesture_id using a StreamSampleManager and a GradientThresholdMiddleware
    """
    print("RECORDING NEW SAMPLES")

    gestures = []
    __saved_gesture = None

    # Create the target directory if it doesn't exists
    if not os.path.exists(target_dir):
        os.makedirs(target_dir)

    # If the gesture id is made of multiple gestures separated by a comma, extract them
    if ',' in gesture_id:
        gestures = map(lambda x: x.strip(), gesture_id.split(","))
    else:
        gestures = [gesture_id]

    print("SERIAL PORT:", port)
    print("TARGET DIRECTORY:", target_dir)
    print("GESTURE ID:", gesture_id)
    print("THRESHOLD:", threshold)

    # Create the SerialDataReader
    sdr = SerialDataReader(port, expected_axis=expected_axis, verbose=False)

    # Create the SampleManager
    manager = StreamSampleManager(step=20, window=20)

    # Attach the manager
    sdr.attach_manager(manager)

    # Create a threshold middleware
    middleware = GradientThresholdMiddleware(verbose=False, threshold=threshold, sample_group_delay=5, group=True)

    # Attach the middleware
    manager.attach_receiver(middleware)

    lfmiddleware = LengthThresholdMiddleware(verbose=True, min_len=180, max_len=600)
    middleware.attach_receiver(lfmiddleware)

    # Also plot the sample
    plotter_mid = PlotterMiddleware()
    middleware.attach_receiver(plotter_mid)

    # Create the gesture chooser
    gesture_chooser = RandomGestureChooser(gestures)

    # Create a FileGestureRecorder
    recorder = FileGestureRecorder(target_dir=target_dir, forced_gesture_id=gesture_chooser, verbose=True)

    # Attach the recorder to the manager
    lfmiddleware.attach_receiver(recorder)

    # Open the serial connection
    print("Opening serial port...")
    sdr.open()
    print("Opened!")

    print("To exit the loop, press Ctrl+C.")

    # Start the main loop
    sdr.mainloop()


def record_new_samples_piezo(port, gesture_id, target_dir, threshold=5):
    """
    Used to record new samples for the specified gesture_id using a StreamSampleManager and a GradientThresholdMiddleware
    """
    print("RECORDING NEW SAMPLES")

    # Create the target directory if it doesn't exists
    if not os.path.exists(target_dir):
        os.makedirs(target_dir)

    print("SERIAL PORT:", port)
    print("TARGET DIRECTORY:", target_dir)
    print("GESTURE ID:", gesture_id)
    print("THRESHOLD:", threshold)

    # Create the SerialDataReader
    sdr = SerialDataReader(port, expected_axis=1, baud_rate=74880, verbose=False)

    # Create the SampleManager
    manager = StreamSampleManager(step=10, window=10)

    # Attach the manager
    sdr.attach_manager(manager)

    # Create a threshold middleware
    middleware = GradientThresholdMiddleware(verbose=False, threshold=threshold, sample_group_delay=20, group=True)

    # Attach the middleware
    manager.attach_receiver(middleware)

    # Also plot the sample
    plotter_mid = PlotterMiddleware()
    middleware.attach_receiver(plotter_mid)

    # Create a FileGestureRecorder
    recorder = FileGestureRecorder(target_dir=target_dir, forced_gesture_id=gesture_id, verbose=True)

    # Attach the recorder to the manager
    middleware.attach_receiver(recorder)

    # Open the serial connection
    print("Opening serial port...")
    sdr.open()
    print("Opened!")

    print("To exit the loop, press Ctrl+C.")

    # Start the main loop
    sdr.mainloop()



from pygarl.base import Sample
from pygarl.abstracts import ControlSignal, AbstractSampleManager


class DiscreteSampleManager(AbstractSampleManager):
    def __init__(self, min_sample_length=10):
        # Call the base constructor to initialize buffer and axis
        AbstractSampleManager.__init__(self)

        self.min_sample_length = min_sample_length  # Minimum number of frames for a valid Sample

    def start_sample(self):
        """
        Called when a Start signal is received
        """
        # Empty the buffer
        self.buffer = []

    def end_sample(self):
        """
        Called when a Stop Signal is received, package the sample and notify the receivers
        """
        self.package_sample()

    def receive_signal(self, signal):
        """
        Called from a DataReader when a signal is received
        """
        # Call the appropriate method based on the received signal
        if signal == ControlSignal.START:
            # When a START signal is received, start a new sample
            self.start_sample()
        elif signal == ControlSignal.STOP:
            # When a STOP signal is received, end the sample
            self.end_sample()

    def receive_data(self, data):
        """
        Called from a DataReader when new data is available
        """
        # Add the current data frame to the buffer
        self.buffer.append(data)

    def package_sample(self):
        """
        Package the sample with the data in the buffer and notify all the attached receivers
        """
        # Notify the receivers only if the sample length is greater than the minimum
        if len(self.buffer) >= self.min_sample_length:
            # Create a sample with the buffer data
            sample = Sample(data=self.buffer)
            # Notify all the attached receivers
            self.notify_receivers(sample)


class StreamSampleManager(AbstractSampleManager):
    # TODO: Documentation
    def __init__(self, window=20, step=10):
        # Call the base constructor to initialize buffer and axis
        AbstractSampleManager.__init__(self)

        # Step size must be lower or equal to the window size, if not, raise an exception
        if step > window:
            raise ValueError("Step size must be lower or equal to the window size.")

        self.window = window  # Determines the size of the window and the sample
        self.step = step  # How much frames should pass between sample packaging

    def end_sample(self):
        """
        Prematurely end the sample, even if the window size has not been reached
        """
        self.package_sample()

    def receive_signal(self, signal):
        """
        Called from a DataReader when a signal is received
        If the STOP signal is received, prematurely end the sample, even if the window size
        has not been reached
        """
        # Call the appropriate method based on the received signal
        if signal == ControlSignal.STOP:
            # When a STOP signal is received, end the sample
            self.end_sample()

    def receive_data(self, data):
        """
        Called from a DataReader when new data is available
        """
        # Add the current data frame to the buffer
        self.buffer.append(data)

        # If the window size has been reached by the buffer
        if len(self.buffer) >= self.window:
            # Package the buffer in a sample and notify the receivers
            self.package_sample()

            # Shift the buffer to the left, deleting the first "step" frames.
            # Example: with a step = 2
            # BUFFER: 1 2 3 4
            # LEFT SHIFT ( N elements, with N = step )
            # BUFFER: 3 4
            self.buffer = self.buffer[self.step:]

    def package_sample(self):
        """
        Package the sample with the data in the buffer and notify all the attached receivers
        """
        # Create a sample with the buffer data
        sample = Sample(data=self.buffer)
        # Notify all the attached receivers
        self.notify_receivers(sample)



from __future__ import print_function
import random


class RandomGestureChooser(object):
    def __init__(self, gestures):
        if isinstance(gestures, list):
            self.gestures = gestures
        else:
            self.gestures = [gestures]

        self.next_gesture = random.choice(gestures)
        print("NEXT GESTURE ", self.next_gesture)

    def get_gesture(self):
        gesture = self.next_gesture
        self.next_gesture = random.choice(self.gestures)
        print("NEXT GESTURE: ", self.next_gesture)
        return gesture



from pygarl.middlewares import GradientThresholdMiddleware, PlotterMiddleware, FFTMiddleware
from pygarl.mocks import VerboseMiddleware
from pygarl.data_readers import SerialDataReader
from pygarl.sample_managers import DiscreteSampleManager, StreamSampleManager

# This example uses a SerialDataReader to read data from a serial port
# uses a StreamSampleManager to package samples and a GradientThresholdMiddleware


def run_example(*args, **kwargs):
    # Create the SerialDataReader
    sdr = SerialDataReader(kwargs['port'], expected_axis=1, baud_rate=74880, verbose=False)

    # Create the SampleManager
    manager = StreamSampleManager(step=10, window=10)

    # Attach the manager
    sdr.attach_manager(manager)

    # Create a threshold middleware
    middleware = GradientThresholdMiddleware(verbose=True, threshold=5, sample_group_delay=20, group=True)

    # Attach the middleware
    manager.attach_receiver(middleware)

    # Also plot the sample
    plotter_mid = PlotterMiddleware()
    middleware.attach_receiver(plotter_mid)

    # Open the serial connection
    sdr.open()
    print("Opened!")

    # Start the main loop
    sdr.mainloop()



import unittest
import os
import shutil
import scipy as sp
from pygarl.abstracts import *
from pygarl.recorders import *
from pygarl.mocks import *
from pygarl.base import *


# To execute tests, go to the project main directory and type:
# python -m unittest discover


class FileGestureRecorderTestCase(unittest.TestCase):
    """
    Tests to check FileGestureRecorder behaviour
    """

    def setUp(self):
        # Create a test directory if it doesn't exists
        if not os.path.exists("test_dir"):
            os.makedirs("test_dir")

        # Create the recorder
        self.recorder = FileGestureRecorder("test_dir")

    def tearDown(self):
        # Destroy the test directory
        shutil.rmtree("test_dir")

        # Destroy the recorder
        self.recorder = None

    def test_invalid_directory_should_raise_exception(self):
        self.assertRaises(ValueError, FileGestureRecorder, "non_existent_directory")

    def test_sample_without_gesture_id_should_raise_exception(self):
        sample = Sample([[]])

        self.assertRaises(ValueError, self.recorder.save_sample, sample)

    def test_save_sample(self):
        sample = MockSample()
        self.assertTrue(self.recorder.save_sample(sample).startswith("TESTSAMPLE_"))

        self.assertTrue("TESTSAMPLE_" in sample.file_path)

    def test_receive_sample(self):
        sample = MockSample()
        self.recorder.receive_sample(sample)

        self.assertTrue("TESTSAMPLE_" in sample.file_path)

    def test_save_sample_filenames_are_unique(self):
        sample = MockSample()
        self.recorder.save_sample(sample)
        filepath1 = sample.file_path

        self.recorder.save_sample(sample)
        filepath2 = sample.file_path

        self.assertTrue(filepath1 != filepath2)

if __name__ == '__main__':
    unittest.main()



from __future__ import print_function
import os
import joblib
import seaborn
import pandas
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap

from pygarl.base import Sample


class ControlSignal:
    """
    Signals that a DataReader sends to a SampleManager
    """
    START = 0
    STOP = 1
    ERROR = 2
    TIMEOUT = 3


class AbstractDataReader(object):
    """
    Represents the Abstraction of the low level data receiver and manages the communication
    with the signal source, as an Arduino or GPIOs pins.
    """

    def __init__(self):
        self.managers = []

    def attach_manager(self, manager):
        """
        Attach a manager to the DataReader, so that when an event occurs the manager is notified
        """
        self.managers.append(manager)

    def detach_manager(self, manager):
        """
        Detach the manager from the DataReader
        """
        self.managers.remove(manager)

    def notify_data(self, data):
        """
        Notify a new set of data to all the attached managers
        :param data: Float array containing the sensor data, every element is an axis reading
        """
        # Cycle through all managers and call their receive_data method, notifying the data event
        for manager in self.managers:
            manager.receive_data(data)

    def notify_signal(self, signal):
        """
        Notify a ControlSignal to all the attached managers
        :param signal: One of the ControlSignal values
        :return: 
        """
        # Cycle through all managers and notify them of the new signal by calling their receive_signal method
        for manager in self.managers:
            manager.receive_signal(signal)

    def open(self):
        """
        Open the connection, not implemented in the abstract class
        """
        raise NotImplementedError("This method is not implemented in the abstract class.")

    def close(self):
        """
        Close the connection, not implemented in the abstract class
        """
        raise NotImplementedError("This method is not implemented in the abstract class.")

    def mainloop(self):
        """
        Not implemented yet, represents the method that endlessly loops listening for events
        """
        raise NotImplementedError("This method is not implemented in the abstract class.")


class Sender(object):
    """
    Manages the transmission of the samples to the attached Receivers
    """

    def __init__(self):
        """
        Initializes the receiver list
        """
        self.receivers = []

    def attach_receiver(self, receiver):
        """
        Attach a receiver to the SampleManager
        :param receiver: Receiver
        """
        self.receivers.append(receiver)

    def detach_receiver(self, receiver):
        """
        Detach the receiver from the SampleManager
        :param receiver:  Receiver
        """
        self.receivers.remove(receiver)

    def notify_receivers(self, sample):
        """
        Notify a sample to all the receivers
        :param sample: Sample
        """
        # Cycle through all receivers, sending them the Sample
        for receiver in self.receivers:
            receiver.receive_sample(sample)


class AbstractSampleManager(Sender):
    """
    Represents the abstraction of a SampleManager, handles the packaging of data into Samples
    The logic involved vary based on the implementation.
    """

    def __init__(self):
        """
        Initializes the buffer
        """
        # Initialize the sender
        Sender.__init__(self)

        self.buffer = []

    def receive_data(self, data):
        raise NotImplementedError("This method is not implemented in the abstract class.")

    def receive_signal(self, signal):
        raise NotImplementedError("This method is not implemented in the abstract class.")

    def package_sample(self):
        raise NotImplementedError("This method is not implemented in the abstract class.")


class Receiver(object):
    """
    Represents the entity that receives a sample and processes it
    """

    def receive_sample(self, sample):
        raise NotImplementedError("This method is not implemented in the abstract class.")


class AbstractGestureRecorder(Receiver):
    """
    Represents the entity that, received a Sample, saves it.
    For example, it can be used to implement a Sample file saver.
    """

    def receive_sample(self, sample):
        raise NotImplementedError("This method is not implemented in the abstract class.")

    def save_sample(self, sample):
        raise NotImplementedError("This method is not implemented in the abstract class.")


class AbstractGesturePredictor(Receiver):
    """
    Received a Sample, tries to predict the corresponding Gesture.
    The logic involved depends on the implementation.
    """

    def __init__(self):
        self.callbacks = []

    def attach_callback_manager(self, manager):
        """
        Attach a CallbackManager to the GesturePredictor
        """
        self.callbacks.append(manager)

    def detach_callback_manager(self, manager):
        """
        Detach the CallbackManager from the GesturePredictor
        """
        self.callbacks.remove(manager)

    def notify_callbacks(self, gesture_id):
        """
        Notify the gesture_id to all the attached CallbackManagers
        """
        # Cycle through all the CallbackManagers and notify the gesture
        for callback in self.callbacks:
            callback.receive_gesture(gesture_id)

    def receive_sample(self, sample):
        """
        Receive the sample, try to predict the correct gesture and notify all the callbacks
        """
        # Predict the gesture
        predicted_gesture = self.predict(sample)
        # Notify all the callbacks
        self.notify_callbacks(predicted_gesture)

    def predict(self, sample):
        """
        Receive the sample and try to predict the correct gesture
        Return the gesture id
        """
        raise NotImplementedError("This method is not implemented in the abstract class.")


class AbstractMiddleware(Sender, Receiver):
    """
    Represents the entity that, after receiving a sample from a Sender, processes it and sends it to another Receiver
    """

    def process_sample(self, sample):
        """
        Receive the sample, process it and then return it.
        If None is returned, Sample will be suppressed and will not be notified to the receivers.
        The logic vary with the implementation.
        """
        # In this case, return the passed sample without processing it
        return sample

    def receive_sample(self, sample):
        """
        Receive a sample, process it and then notify the result to all the attached receivers
        """
        # Process the sample
        processed_sample = self.process_sample(sample)

        # If the processed_sample is None, don't send a notification to the receivers
        # In this way, a middleware can suppress a sample if needed.
        if processed_sample is not None:
            # Send the processed sample to all the attached receivers
            self.notify_receivers(processed_sample)


class AbstractClassifier(object):
    """
    Represents an entity that takes a directory containing a set of samples
    and creates a model that can be used to predict at which gesture a sample
    belongs to.
    """
    def __init__(self, dataset_path=None, model_path=None, verbose=False, autonormalize=False, autoscale_size=None,
                       middlewares=[]):
        """
        :param dataset_path: path to the directory containing the samples dataset.
        :param model_path: path to a saved model file.
        :param verbose: if true, the classifier will print the training progress.
        :param autonormalize: if true, the classifier will normalize the samples.
        :param autoscale_size: an integer, if set, the classifier will interpolate the samples frames
                               to scale them to the number specified by this parameter.
        :param middlewares: a list of Middlewares that will be applied to each loaded sample before
                            loading the data. They are applied before the auto-normalization.
        """
        # Dataset_path and model_path must be mutually exclusive and can't be both defined.
        # That's because dataset_path is used in the training phase, while
        # model_path is used when loading an existing model.
        # Those are very different moments and should be handled differently
        if model_path is not None and dataset_path is not None:
            # If both are defined, raise an exception
            raise ValueError("dataset_path and model_path must not be defined together.")

        # One of them must be defined, if not, raise an exception
        if model_path is None and dataset_path is None:
            raise ValueError("You must define model_path or dataset_path")

        self.dataset_path = dataset_path
        self.model_path = model_path
        self.verbose = verbose
        self.autonormalize = autonormalize
        self.autoscale_size = autoscale_size
        self.middlewares = middlewares

        # This is initially false and becomes true only when a valid model is ready
        # That could happen when a model is trained or loaded
        # If this variable is false, no prediction can be made
        self.is_trained = False

        # Contains the list of all gesture_ids, the index represents the internal id
        self.gestures = []

        # Contains the list of filenames of all samples contained in the dataset
        # Initially it is none and must be populated with load_samples_filenames
        self.samples_filenames = None

        # After the classifier has been trained, the confusion matrix is generated
        self.confusion_matrix = None

    def load(self):
        """
        Used to load the samples data and ids if dataset_path is set
        or to load the model if model_path is set.
        """
        # If the dataset_path is set, load the samples
        if self.dataset_path is not None:
            # Load the data needed
            self.load_samples_filenames()
            self.load_gestures_ids()
            self.load_samples_data()

        elif self.model_path is not None:  # If model_path is set, load the model
            # Load the model
            self.load_from_file()

    def load_samples_filenames(self):
        """
        Load the samples' filenames list contained in the given dataset
        """
        # If the dataset is not defined, samples can't be loaded so raise an exception
        if self.dataset_path is None:
            raise ValueError("dataset_path must be defined to load samples filenames.")

        # Get the list of files contained in the dataset_path
        self.samples_filenames = [f for f in os.listdir(self.dataset_path)
                                  if os.path.isfile(os.path.join(self.dataset_path, f))]

        return self.samples_filenames

    def get_internal_id_from_gesture_id(self, gesture_id):
        """
        Return the gesture internal_id. If the gesture_id is not found, raises a ValueError
        :param gesture_id: a string containing a gesture_id
        :return: the internal id
        """

        # Return the internal_id of the given gesture_id
        # If the gesture_id is not found, raises a ValueError
        return self.gestures.index(gesture_id)

    @staticmethod
    def get_gesture_id_from_filename(filename):
        """
        Return the gesture_id from the given filename
        :param filename: Name of the file, without the path
        :return: the gesture_id as string
        """
        # Get the gesture out of the filename by taking the string before the first underscore _
        gesture_id = filename.split("_")[0]

        return gesture_id

    def load_gestures_ids(self):
        """
        Load the gestures' ids of the samples contained in the given dataset
        """
        # If the dataset is not defined, gestures can't be loaded so raise an exception
        if self.dataset_path is None:
            raise ValueError("dataset_path must be defined to load the gestures.")

        # Samples' filenames must have been loaded before calling this method
        if self.samples_filenames is None:
            raise ValueError("samples_filenames must be loaded before calling this method. "
                             "That can be done using load_samples_filenames()")

        # Cycle through all file names
        for f in self.samples_filenames:
            # Get the gesture out of the filename
            gesture_id = AbstractClassifier.get_gesture_id_from_filename(f)

            # If gesture_id is not already present in the dictionary
            if gesture_id not in self.gestures:
                # Add the new gesture_id to the list
                self.gestures.append(gesture_id)

        return self.gestures

    def load_samples_data(self):
        """
        Loads the samples data by cycling through each of them and calling the
        implementation-specific "load_sample_data" for each of them
        """
        # If the dataset is not defined, gestures can't be loaded so raise an exception
        if self.dataset_path is None:
            raise ValueError("dataset_path must be defined to load the gestures.")

        # Samples' filenames must have been loaded before calling this method
        if self.samples_filenames is None:
            raise ValueError("samples_filenames must be loaded before calling this method. "
                             "That can be done using load_samples_filenames()")

        # Cycle through all file names
        for f in self.samples_filenames:
            # Generate the complete sample path
            complete_path = os.path.join(self.dataset_path, f)

            # Load the sample
            sample = Sample.load_from_file(complete_path)

            # Apply all the middlewares
            for middleware in self.middlewares:
                sample = middleware.process_sample(sample)

            # If autonormalize is set, normalize the sample's frames
            if self.autonormalize:
                sample.normalize_frames()

            # If autoscale_size is set, scale the number of frames to the specified value
            if self.autoscale_size is not None:
                sample.scale_frames(n_frames=self.autoscale_size)

            # Call the implementation-specific load_sample_data method
            self.load_sample_data(sample)

    def predict(self, sample):
        """
        Return the gesture id associated with the given sample ( using a prediction algorithm ).
        IMPORTANT: to customize the prediction algorithm, you must override the 
        "predict_sample" method, not this one.
        
        :param sample: sample used to predict the gesture
        :return: a string containing the "gesture_id" of the sample
        """
        # The model must be trained before making a prediction, if not, raise an exception
        if not self.is_trained:
            raise ValueError("The model must be trained before making a prediction")

        # Apply all the middlewares
        for middleware in self.middlewares:
            sample = middleware.process_sample(sample)

        # If autonormalize is set, normalize the sample's frames
        if self.autonormalize:
            sample.normalize_frames()

        # If autoscale_size is set, scale the number of frames to the specified value
        if self.autoscale_size is not None:
            sample.scale_frames(n_frames=self.autoscale_size)

        # Pass the sample to the inner prediction function
        return self.predict_sample(sample)

    def load_sample_data(self, sample):
        """
        Called for each sample in the dataset, should handle the manipulation of data
        used later to train the classifier.
        """
        raise NotImplementedError("This method is not implemented in the abstract class.")

    def print_info(self):
        """
        Print the classifier info.
        """
        print("AUTONORMALIZE:", self.autonormalize)
        print("AUTOSCALE_SIZE:", self.autoscale_size)
        print("GESTURES:", self.gestures)
        print("MIDDLEWARES:", self.middlewares)

    def get_attributes(self):
        """
        Return a dictionary containing the needed attributes to save the classifier
        """
        return {'verbose': self.verbose, 'autonormalize': self.autonormalize,
                'autoscale_size': self.autoscale_size, 'gestures': self.gestures,
                'middlewares': self.middlewares}

    def load_attributes(self, attributes):
        """
        Load the specified attributes in the classifier.
        :param attributes: a dictionary containing the attributes
        """
        # Set the parameters
        self.verbose = attributes['verbose']
        self.autonormalize = attributes['autonormalize']
        self.autoscale_size = attributes['autoscale_size']
        self.gestures = attributes['gestures']
        self.middlewares = attributes['middlewares']

    def predict_sample(self, sample):
        raise NotImplementedError("This method is not implemented in the abstract class.")

    def train_model(self):
        raise NotImplementedError("This method is not implemented in the abstract class.")

    def save_model(self, model_path):
        """
        Save the model to the specified path.
        Note: the model must be trained before saving.

        :param model_path: path of the output model file 
        """
        # If the model is not trained, raise an exception
        if not self.is_trained:
            raise ValueError("The model must be trained before saving it.")

        # Get the attributes that must be saved
        output_data = self.get_attributes()

        # Dump the model to a file
        joblib.dump(output_data, model_path)

    def load_from_file(self):
        """
        Load a previously saved model from a file, specified by the
        self.model_path parameter, specified in the constructor
        """
        # Load the data from the model file
        input_data = joblib.load(self.model_path)

        # Populate the class attributes
        self.load_attributes(input_data)

        # Set the model as trained
        self.is_trained = True

    def plot_confusion_matrix(self):
        """
        Plot the confusion matrix using Matplotlib and seaborn.
        The classifier must have been trained beforehand, if this is not true,
        this method will raise a ValueError
        """
        if self.confusion_matrix is None:
            raise ValueError("The confusion matrix is not yet defined, you have to train your classifier first!")

        # Create the dataframe for the confusion matrix
        df_cm = pandas.DataFrame(self.confusion_matrix,
                                 index=self.gestures,
                                 columns=self.gestures)

        # Create the HeatMap
        plt.figure(figsize=(10, 7))
        seaborn.set(font_scale=1.4)
        cmap = ListedColormap(seaborn.color_palette("PuBu", 10))
        seaborn.heatmap(df_cm, annot=True, cmap=cmap)

        # Show the Plot
        plt.show()



import sys

from pygarl.base import Sample
from pygarl.middlewares import PlotterMiddleware, GradientThresholdMiddleware, TrimmerMiddleware
from pygarl.mocks import VerboseMiddleware
from pygarl.data_readers import SerialDataReader
from pygarl.sample_managers import DiscreteSampleManager

# This example plots a sample from file


def run_example(*args, **kwargs):
    sample = Sample.load_from_file(args[0])

    tm = TrimmerMiddleware(threshold=300)

    sample = tm.process_sample(sample)

    sample.plot()



from pygarl.middlewares import PlotterMiddleware
from pygarl.mocks import VerboseMiddleware
from pygarl.data_readers import SerialDataReader
from pygarl.sample_managers import DiscreteSampleManager

# This example uses a SerialDataReader to read data from a serial port
# and plot the recorded sample


def run_example(*args, **kwargs):
    # Create the SerialDataReader
    sdr = SerialDataReader(kwargs['port'], expected_axis=6, verbose=False)

    # Create the SampleManager
    manager = DiscreteSampleManager()

    # Attach the manager
    sdr.attach_manager(manager)

    # Create the VerboseMiddleware that prints the received sample
    middleware = PlotterMiddleware()

    # Attach the middleware
    manager.attach_receiver(middleware)

    # Open the serial connection
    sdr.open()
    print("Opened!")

    # Start the main loop
    sdr.mainloop()



from __future__ import print_function
from pygarl.abstracts import *
from pygarl.base import Sample


class MockSampleManager(AbstractSampleManager):
    """
    Mock implementation of the sample manager, used for tests
    """

    def __init__(self):
        AbstractSampleManager.__init__(self)

        self.received_data = None
        self.received_signal = None

    def receive_data(self, data):
        self.received_data = data

    def receive_signal(self, signal):
        self.received_signal = signal

    def package_sample(self):
        pass


class VerboseTestSampleManager(AbstractSampleManager):
    """
    Used to print the received values from a DataReader
    """

    def __init__(self):
        AbstractSampleManager.__init__(self)

    def receive_data(self, data):
        print("DATA:", data)

    def receive_signal(self, signal):
        print("SIGNAL:", signal)

    def package_sample(self):
        pass


class VerboseMiddleware(AbstractMiddleware):
    """
    Used to print the received sample
    """
    def __init__(self, verbose=True):
        AbstractMiddleware.__init__(self)
        self.counter = 0
        self.verbose = verbose

    def process_sample(self, sample):
        if self.verbose:
            print("SAMPLE "+str(self.counter))
            print(sample)
            self.counter += 1
        return sample


class MockReceiver(Receiver):
    """
    Mock implementation of the receiver, used for tests
    """
    def __init__(self):
        Receiver.__init__(self)

        self.received_sample = None

    def receive_sample(self, sample):
        self.received_sample = sample


class MockSample(Sample):
    """
    Mock implementation of Sample, used for tests
    """
    def __init__(self):
        Sample.__init__(self, [[]], gesture_id="TESTSAMPLE")

        self.file_path = None

    def save_to_file(self, file_path):
        self.file_path = file_path


class MockFunctionReceiver(object):
    """
    Used to test the callback manager
    """
    def __init__(self):
        self.received = False

    def receive(self, gesture_id):
        """
        When this function is called, self.received becomes true 
        """
        self.received = True


class MockFunctionCounter:
    """
    Used to test if a function was called, and how many times
    """
    def __init__(self):
        self.counter = 0

    def callback(self, *args):
        self.counter += 1


class MockCallbackManager(object):
    """
    Used to test the callback manager
    """
    def __init__(self):
        self.received_gesture = None

    def receive_gesture(self, gesture_id):
        """
        When this function is called, set self.received_gesture
        """
        self.received_gesture = gesture_id



import unittest

import shutil

from pygarl.abstracts import *
from pygarl.mocks import *
from pygarl.base import *

# To execute tests, go to the project main directory and type:
# python -m unittest discover


class AbstractDataReaderTestCase(unittest.TestCase):
    """
    Tests to check AbstractDataReader consistency
    """

    def setUp(self):
        # Initialize an AbstractDataReader
        self.abstract_data_reader = AbstractDataReader()

    def tearDown(self):
        # Destroy the AbstractDataReader
        self.abstract_data_reader = None

    def test_manager_attached_correctly(self):
        sample_manager = AbstractSampleManager()

        # Attach the manager
        self.abstract_data_reader.attach_manager(sample_manager)

        # Check if the manager is attached
        self.assertIn(sample_manager, self.abstract_data_reader.managers)

    def test_manager_detached_correctly(self):
        sample_manager = AbstractSampleManager()

        # Attach the manager
        self.abstract_data_reader.attach_manager(sample_manager)

        # Check if the manager is attached
        self.assertIn(sample_manager, self.abstract_data_reader.managers)

        # Detach the manager
        self.abstract_data_reader.detach_manager(sample_manager)

        # Check that the manager is not attached anymore
        self.assertNotIn(sample_manager, self.abstract_data_reader.managers)

    def test_notify_data(self):
        sample_manager = MockSampleManager()

        # Attach the manager
        self.abstract_data_reader.attach_manager(sample_manager)

        # Check that the received data is initially None
        self.assertIsNone(sample_manager.received_data)

        # Notify the data by sending True
        self.abstract_data_reader.notify_data(True)

        # Check that the received data is not none anymore
        self.assertIsNotNone(sample_manager.received_data)

        # Check that the received data is True
        self.assertTrue(sample_manager.received_data)

    def test_notify_signal(self):
        sample_manager = MockSampleManager()

        # Attach the manager
        self.abstract_data_reader.attach_manager(sample_manager)

        # Check that the received signal is initially None
        self.assertIsNone(sample_manager.received_signal)

        # Notify the signal by sending START
        self.abstract_data_reader.notify_signal(ControlSignal.START)

        # Check that the received signal is not none anymore
        self.assertIsNotNone(sample_manager.received_signal)

        # Check that the received data is START
        self.assertEqual(sample_manager.received_signal, ControlSignal.START)

    def test_mainloop_not_implemented(self):
        # The mainloop function must be abstract
        self.assertRaises(NotImplementedError, self.abstract_data_reader.mainloop)


class SenderTestCase(unittest.TestCase):
    """
    Tests to check Sender consistency
    """
    def setUp(self):
        # Initialize a Sender
        self.sender = Sender()

    def tearDown(self):
        # Destroy the Sender
        self.sender = None

    def test_receiver_attached_correctly(self):
        receiver = Receiver()
        # Receiver must not be already attached
        self.assertNotIn(receiver, self.sender.receivers)
        # Attach the receiver
        self.sender.attach_receiver(receiver)
        # Check that the receiver is attached
        self.assertIn(receiver, self.sender.receivers)

    def test_receiver_detached_correctly(self):
        receiver = Receiver()
        # Attach the receiver
        self.sender.attach_receiver(receiver)
        # Check that the receiver is attached
        self.assertIn(receiver, self.sender.receivers)
        # Detach the receiver
        self.sender.detach_receiver(receiver)
        # Check that the receiver is detached
        self.assertNotIn(receiver, self.sender.receivers)

    def test_notify_receivers(self):
        sample = Sample([[]])
        receiver = MockReceiver()
        # Attach the receiver
        self.sender.attach_receiver(receiver)
        # Initially the received sample must be none
        self.assertIsNone(receiver.received_sample)
        # Notify the receivers
        self.sender.notify_receivers(sample)
        # The received sample must be equal to the sent one
        self.assertEqual(receiver.received_sample, sample)


class AbstractSampleManagerTestCase(unittest.TestCase):
    """
    Tests to check AbstractSampleManager consistency
    """

    def setUp(self):
        # Initialize an AbstractSampleManager
        self.abstract_sample_manager = AbstractSampleManager()

    def tearDown(self):
        # Destroy the AbstractSampleManager
        self.abstract_sample_manager = None

    def test_receiver_attached_correctly(self):
        receiver = Receiver()
        # Receiver must not be already attached
        self.assertNotIn(receiver, self.abstract_sample_manager.receivers)
        # Attach the receiver
        self.abstract_sample_manager.attach_receiver(receiver)
        # Check that the receiver is attached
        self.assertIn(receiver, self.abstract_sample_manager.receivers)

    def test_receiver_detached_correctly(self):
        receiver = Receiver()
        # Attach the receiver
        self.abstract_sample_manager.attach_receiver(receiver)
        # Check that the receiver is attached
        self.assertIn(receiver, self.abstract_sample_manager.receivers)
        # Detach the receiver
        self.abstract_sample_manager.detach_receiver(receiver)
        # Check that the receiver is detached
        self.assertNotIn(receiver, self.abstract_sample_manager.receivers)

    def test_notify_receivers(self):
        sample = Sample([[]])
        receiver = MockReceiver()
        # Attach the receiver
        self.abstract_sample_manager.attach_receiver(receiver)
        # Initially the received sample must be none
        self.assertIsNone(receiver.received_sample)
        # Notify the receivers
        self.abstract_sample_manager.notify_receivers(sample)
        # The received sample must be equal to the sent one
        self.assertEqual(receiver.received_sample, sample)

    def test_receive_data_not_implemented(self):
        # The function must be abstract
        self.assertRaises(NotImplementedError, self.abstract_sample_manager.receive_data, None)

    def test_receive_signal_not_implemented(self):
        # The function must be abstract
        self.assertRaises(NotImplementedError, self.abstract_sample_manager.receive_signal, None)

    def test_package_sample_not_implemented(self):
        # The function must be abstract
        self.assertRaises(NotImplementedError, self.abstract_sample_manager.package_sample)


class AbstractGesturePredictorTestCase(unittest.TestCase):
    """
    Tests to check AbstractGesturePredictor behaviour
    """

    def setUp(self):
        # Initialize an AbstractGesturePredictor
        self.abstract_gesture_predictor = AbstractGesturePredictor()

    def tearDown(self):
        # Destroy the AbstractGesturePredictor
        self.abstract_gesture_predictor = None

    def test_callback_manager_attached_correctly(self):
        callback_mg = CallbackManager()
        # Callback Manager must not be already attached
        self.assertNotIn(callback_mg, self.abstract_gesture_predictor.callbacks)
        # Attach the Callback Manager
        self.abstract_gesture_predictor.attach_callback_manager(callback_mg)
        # Check that the callback manager is attached
        self.assertIn(callback_mg, self.abstract_gesture_predictor.callbacks)

    def test_callback_manager_detached_correctly(self):
        callback_mg = CallbackManager()

        # Attach the Callback Manager
        self.abstract_gesture_predictor.attach_callback_manager(callback_mg)
        # Check that the callback manager is attached
        self.assertIn(callback_mg, self.abstract_gesture_predictor.callbacks)
        # Detach the Callback Manager
        self.abstract_gesture_predictor.detach_callback_manager(callback_mg)
        # Callback Manager must not be already attached
        self.assertNotIn(callback_mg, self.abstract_gesture_predictor.callbacks)

    def test_notify_callbacks(self):
        callback_mg = MockCallbackManager()

        self.abstract_gesture_predictor.attach_callback_manager(callback_mg)
        # Initially the received gesture is None
        self.assertIsNone(callback_mg.received_gesture)
        # Notify the gesture
        self.abstract_gesture_predictor.notify_callbacks("TEST")
        # Make sure the gesture has been received
        self.assertEqual(callback_mg.received_gesture, "TEST")

    def test_predict_not_implemented(self):
        sample = Sample([[]])
        # The function must be abstract
        self.assertRaises(NotImplementedError, self.abstract_gesture_predictor.predict, sample)


class AbstractMiddlewareTestCase(unittest.TestCase):
    """
    Tests to check AbstractMiddleware behaviour
    """

    def setUp(self):
        # Initialize an AbstractMiddleware
        self.abstract_middleware = AbstractMiddleware()

    def tearDown(self):
        # Destroy the AbstractMiddleware
        self.abstract_middleware = None

    def test_notify_receivers(self):
        sample = Sample([[]])
        receiver = MockReceiver()
        # Attach the receiver
        self.abstract_middleware.attach_receiver(receiver)
        # Initially the received sample must be none
        self.assertIsNone(receiver.received_sample)
        # Notify the receivers
        self.abstract_middleware.notify_receivers(sample)
        # The received sample must be equal to the sent one
        self.assertEqual(receiver.received_sample, sample)

    def test_receive_sample_and_notify_receivers(self):
        sample = Sample([[]])
        receiver = MockReceiver()
        # Attach the receiver
        self.abstract_middleware.attach_receiver(receiver)
        # Initially the received sample must be none
        self.assertIsNone(receiver.received_sample)
        # Send the sample to the receiver
        self.abstract_middleware.receive_sample(sample)
        # The received sample must be equal to the sent one
        self.assertEqual(receiver.received_sample, sample)

    def test_process_sample_should_not_process(self):
        sample = Sample([[]])
        # The function should return the sample without processing it ( in the abstract class )
        self.assertEqual(sample, self.abstract_middleware.process_sample(sample))


class AbstractClassifierTestCase(unittest.TestCase):
    """
    Tests to check AbstractClassifier consistency
    """

    def create_a_file(self, filename, content='{"gesture_id": "TESTSAMPLE", "data": [[1, 2, 3], [4, 5, 6]]}'):
        """
        Create a file, used for tests
        """
        with open(os.path.join("test_dir_abstract_classifier", filename), 'w') as output_file:
            output_file.write(content)

    def setUp(self):
        # Create a test directory if it doesn't exists
        if not os.path.exists("test_dir_abstract_classifier"):
            os.makedirs("test_dir_abstract_classifier")

        # Create a bunch of files
        self.create_a_file("id1_0_0.txt")
        self.create_a_file("id1_1_0.txt")
        self.create_a_file("id2_1_0.txt")
        self.create_a_file("id5_1_0.txt")
        self.create_a_file("id1_2_0.txt")

        # Initialize an AbstractClassifier
        self.classifier = AbstractClassifier(dataset_path="test_dir_abstract_classifier")

    def tearDown(self):
        # Destroy the test directory
        shutil.rmtree("test_dir_abstract_classifier")

        # Destroy the AbstractClassifier
        self.classifier = None

    def test_dataset_and_model_both_defined_should_raise_error(self):
        self.assertRaises(ValueError, AbstractClassifier, dataset_path="1", model_path="2")

    def test_at_least_one_dataset_or_model_must_be_defined_should_raise_error(self):
        self.assertRaises(ValueError, AbstractClassifier)

    def test_load_gestures_ids_should_fail_if_called_before_loading_samples_filenames(self):
        self.assertRaises(ValueError, self.classifier.load_gestures_ids)

    def test_load_gestures_ids_should_fail_if_dataset_path_is_not_defined(self):
        # Reset the dataset path
        self.classifier.dataset_path = None
        self.assertRaises(ValueError, self.classifier.load_gestures_ids)

    def test_load_samples_filenames_should_fail_if_dataset_path_is_not_defined(self):
        # Reset the dataset path
        self.classifier.dataset_path = None
        self.assertRaises(ValueError, self.classifier.load_samples_filenames)

    def test_load_samples_filenames(self):
        filenames = self.classifier.load_samples_filenames()

        self.assertIn("id1_0_0.txt", filenames)
        self.assertIn("id1_1_0.txt", filenames)
        self.assertIn("id2_1_0.txt", filenames)
        self.assertIn("id5_1_0.txt", filenames)
        self.assertIn("id1_2_0.txt", filenames)

    def test_load_gestures_ids(self):
        self.classifier.load_samples_filenames()

        ids = self.classifier.load_gestures_ids()

        self.assertIn("id1", ids)
        self.assertIn("id2", ids)
        self.assertIn("id5", ids)

    def test_get_gesture_id_from_filename(self):
        self.assertEqual("id1", AbstractClassifier.get_gesture_id_from_filename("id1_0_0"))

    def test_load_samples_data_should_fail_if_called_before_loading_samples_filenames(self):
        self.assertRaises(ValueError, self.classifier.load_samples_data)

    def test_load_samples_data_should_fail_if_dataset_path_is_not_defined(self):
        # Reset the dataset path
        self.classifier.dataset_path = None
        self.assertRaises(ValueError, self.classifier.load_samples_data)

    def test_load_samples_data(self):
        # Substitute the load_sample_data method of the classifier with a Mock
        mock = MockFunctionCounter()

        self.classifier.load_sample_data = mock.callback
        # Counter must be zero initially
        self.assertEqual(mock.counter, 0)

        self.classifier.load_samples_filenames()
        self.classifier.load_samples_data()

        # Counter must be zero initially
        self.assertEqual(mock.counter, len(self.classifier.samples_filenames))

    def test_predict_should_fail_if_the_model_is_not_trained(self):
        self.assertFalse(self.classifier.is_trained)

        self.assertRaises(ValueError, self.classifier.predict, Sample([[]]))

    def test_plot_confusion_matrix_should_fail_before_training(self):
        self.assertRaises(ValueError, self.classifier.plot_confusion_matrix)


if __name__ == '__main__':
    unittest.main()



import serial

import sys

from pygarl.abstracts import AbstractDataReader, ControlSignal


class SerialDataReader(AbstractDataReader):
    """
    Used to get the data needed to make a sample from a serial connection
    """
    def __init__(self, serial_port, baud_rate=38400, timeout=1, expected_axis=6, verbose=False):
        AbstractDataReader.__init__(self)

        self.serial_port = serial_port
        self.baud_rate = baud_rate
        self.timeout = timeout
        self.verbose = verbose
        self.expected_axis = expected_axis

        # Set the serial connection as None initially
        self.serial = None

    def open(self):
        """
        Open the serial connection using the parameters specified in the class constructor
        """
        # Make sure the serial connection is not open, if open raise an exception
        if self.serial is not None:
            raise RuntimeError("The serial connection is already opened")

        # Open the serial connection
        self.serial = serial.Serial(self.serial_port, self.baud_rate, timeout=self.timeout)

    def close(self):
        """
        Close the serial connection
        """
        # Check if the serial connection has been opened, if not raise an exception
        if self.serial is None:
            raise RuntimeError("The serial connection is not open, so it can't be closed.")

        # If the connection was open, close it
        self.serial.close()

        # Destroy the serial object
        self.serial = None

    def mainloop(self):
        """
        Endless loop that waits for data from the serial connection and dispatch events when they occur
        """
        # Enclosed in a try block to intercept a Ctrl+C press
        try:
            # Start the endless loop
            while True:
                # Read a line from the serial connection
                line = self.serial.readline()

                # This line is needed for backward compatibility
                # Converts the string into binary data for python > 3
                if sys.version_info >= (3,):
                    line = line.decode("utf-8")

                # Deleting the new line characters
                line = line.replace("\r\n", "")

                # If verbosity is true, print the received line
                if self.verbose:
                    print(line)

                # Analyze the received data, dispatching the correct event based on the content
                if line == "STARTING BATCH":
                    # Batch started, dispatch the START event
                    self.notify_signal(ControlSignal.START)
                elif line == "CLOSING BATCH":
                    # Batch closed, dispatch the STOP event
                    self.notify_signal(ControlSignal.STOP)
                elif line.startswith("START") and line.endswith("END"):
                    # Data line, parse the data
                    # A data line should have this format
                    # START -36 1968 16060 -108 258 -136 END

                    # Get the values by splitting the line
                    value_list = line.split(" ")

                    # Excluding START and END, there should be at least one value
                    # ( so the total length should be 2 + number of expected axis ).
                    if len(value_list) == (2 + self.expected_axis):
                        # Get the values contained in the list, by removing the first and last element ( START and END )
                        string_values = value_list[1:-1]

                        # Convert the values from string to float
                        values = []

                        # If true, the data line will be invalidated
                        has_errors = False

                        for value in string_values:
                            # Make sure that the character can be converted to float
                            try:
                                fvalue = float(value)  # Convert the string
                                values.append(fvalue)  # Add to the values array
                            except ValueError:  # Conversion error, mark the data line as wrong
                                has_errors = True

                        # Before sending the values, make sure they are error free
                        if not has_errors:
                            # Dispatch the DATA event, sending the values
                            self.notify_data(values)
                        else:
                            # An error occurred, dispatch the ERROR event
                            self.notify_signal(ControlSignal.ERROR)
                    else:  # An error occurred, dispatch the ERROR event
                        self.notify_signal(ControlSignal.ERROR)
                elif line == "":  # This could be a timeout
                    # Dispatch the TIMEOUT event
                    self.notify_signal(ControlSignal.TIMEOUT)
                else:  # This must be an error
                    # Dispatch the ERROR event
                    self.notify_signal(ControlSignal.ERROR)
        except KeyboardInterrupt:  # When Ctrl+C is pressed, the loop terminates
            print('CLOSED MAINLOOP!')


class FileDataReader(AbstractDataReader):
    """
    Used to simulate a data connection by reading the values from a file
    """
    def __init__(self, file_path, verbose=False):
        AbstractDataReader.__init__(self)

        self.file_path = file_path
        self.verbose = verbose

        # Set the file as None initially
        self.file = None

    def open(self):
        """
        Open the file
        """
        # Make sure the file is not open, if open raise an exception
        if self.file is not None:
            raise RuntimeError("The file is already opened")

        # Open the file
        self.file = open(self.file_path, "rb")

    def close(self):
        """
        Close the file
        """
        # Check if the file has been opened, if not raise an exception
        if self.file is None:
            raise RuntimeError("The file is not open, so it can't be closed.")

        # Close the file
        self.file.close()

        # Destroy the file object
        self.file = None

    def mainloop(self):
        """
        Loop that reads file lines and dispatch events when they occur
        """
        # Enclosed in a try block to intercept a Ctrl+C press
        try:
            # Start the loop
            while True:
                # Read a line from the file
                line = self.file.readline()

                # Check if the file is finished
                if not line:
                    break

                # This line is needed for backward compatibility
                # Converts the string into binary data for python > 3
                if sys.version_info >= (3,):
                    line = line.decode("utf-8")

                # Deleting the new line characters
                line = line.replace("\r\n", "")

                # If verbosity is true, print the received line
                if self.verbose:
                    print(line)

                # Analyze the received data, dispatching the correct event based on the content
                if line == "STARTING BATCH":
                    # Batch started, dispatch the START event
                    self.notify_signal(ControlSignal.START)
                elif line == "CLOSING BATCH":
                    # Batch closed, dispatch the STOP event
                    self.notify_signal(ControlSignal.STOP)
                elif line.startswith("START") and line.endswith("END"):
                    # Data line, parse the data
                    # A data line should have this format
                    # START -36 1968 16060 -108 258 -136 END

                    # Get the values by splitting the line
                    value_list = line.split(" ")

                    # Check the value number
                    if len(value_list) > 2:
                        # Get the values contained in the list, by removing the first and last element ( START and END )
                        string_values = value_list[1:-1]

                        # Convert the values from string to float
                        values = []

                        # If true, the data line will be invalidated
                        has_errors = False

                        for value in string_values:
                            # Make sure that the character can be converted to float
                            try:
                                fvalue = float(value)  # Convert the string
                                values.append(fvalue)  # Add to the values array
                            except ValueError:  # Conversion error, mark the data line as wrong
                                has_errors = True

                        # Before sending the values, make sure they are error free
                        if not has_errors:
                            # Dispatch the DATA event, sending the values
                            self.notify_data(values)
                        else:
                            # An error occurred, dispatch the ERROR event
                            self.notify_signal(ControlSignal.ERROR)
                    else:  # An error occurred, dispatch the ERROR event
                        self.notify_signal(ControlSignal.ERROR)
                elif line == "":  # This could be a timeout
                    # Dispatch the TIMEOUT event
                    self.notify_signal(ControlSignal.TIMEOUT)
                else:  # This must be an error
                    # Dispatch the ERROR event
                    self.notify_signal(ControlSignal.ERROR)
        except KeyboardInterrupt:  # When Ctrl+C is pressed, the loop terminates
            print('CLOSED MAINLOOP!')



from pygarl.data_readers import SerialDataReader
from pygarl.mocks import VerboseTestSampleManager

# This example uses a SerialDataReader to read data from a serial port
# and uses a VerboseTestSampleManager to print the received data and signals


def run_example(*args, **kwargs):
    # Create the SerialDataReader
    sdr = SerialDataReader(kwargs['port'], expected_axis=6, verbose=False)

    # Create a simple SampleManager that only prints the received data and signals
    manager = VerboseTestSampleManager()

    # Attach the manager
    sdr.attach_manager(manager)

    # Open the serial connection
    sdr.open()
    print("Opened!")

    # Start the main loop
    sdr.mainloop()



import click  # Use the click library to provide a CLI interface
import importlib
import os

from pygarl.plugins.plist import list_serial_ports
from pygarl.plugins.plot import plot_sample
from pygarl.plugins.record import record_new_samples, record_new_samples_stream, record_new_samples_piezo
from pygarl.plugins.train import train_svm_classifier, train_mlp_classifier
from pygarl.plugins.sprint import sprint as sprint_func


def get_default_record_directory():
    """
    Used to generate the default user directory for saving new samples
    """
    # Get the user HOME directory
    home = os.path.expanduser("~")

    # Generate the complete path as: $HOME/dataset
    complete_path = os.path.join(home, "dataset")

    return complete_path


@click.group()
def cli():
    pass


@cli.command()
@click.option('--port', '-p', default="COM6", help="Serial Port NAME, for example COM3.")
@click.option('--dir', '-d', default=get_default_record_directory(),
              help="Target directory where samples will be saved.")
@click.option('--gesture', '-g', default="SAMPLE",
              help="Gesture ID of the recorded samples.")
@click.option('--axis', '-a', default=6, help="Number of AXIS in the signal, default 6.")
@click.option('--mode', '-m', default="discrete", help="Recording mode. You can choose between discrete, stream and piezo.")
@click.option('--threshold', '-t', default=40, help="If stream or piezo mode is specified, use this parameter to"
                                                            "regulate the threshold for the input.")
def record(port, dir, gesture, axis, mode, threshold):
    """
    Record new samples and saves them to file
    """
    if mode == "discrete":
        record_new_samples(port=port, gesture_id=gesture, target_dir=dir, expected_axis=axis)
    elif mode == "stream":
        record_new_samples_stream(port=port, gesture_id=gesture, target_dir=dir, expected_axis=axis, threshold=threshold)
    elif mode == "piezo":
        record_new_samples_piezo(port=port, gesture_id=gesture, target_dir=dir, threshold=threshold)


@cli.command()
def plist():
    """
    Prints all the available serial ports
    """
    list_serial_ports()


@cli.command()
@click.option('--port', '-p', default="COM6", help="Serial Port NAME, for example COM3.")
@click.option('--baudrate', '-b', default=9600, help="Serial Port Baudrate, default 9600.")
def sprint(port, baudrate):
    """
    Print the input received from the specified serial port
    """
    sprint_func(port=port, baudrate=baudrate)


@cli.command()
@click.option('--dir', '-d', default=get_default_record_directory(),
              help="Dataset directory where samples are saved.")
@click.option('--classifier', '-c', default="svm",
              help="Classifier used to create a model. Default is SVM. You can use svm, mlp and custom.")
@click.option('--trainer', '-t', default=None,
              help="Load a custom trainer. --classifier custom must be specified.")
@click.argument('output_file')
def train(dir, classifier, output_file, trainer):
    """
    Train a model from a dataset
    """
    # Load the appropriate method based on the specified classifier
    if classifier == "svm":
        train_svm_classifier(dir, output_file)
    elif classifier == "mlp":
        train_mlp_classifier(dir, output_file)
    elif classifier == "custom":
        if trainer is None:
            raise ValueError("If --classifier custom is used, a trainer must be specified")

        # Load the trainer
        trainer = importlib.import_module("pygarl.plugins.trainers." + trainer)

        # Train the classifier
        trainer.train(dir, output_file)
    else:
        raise ValueError("{classifier} is not a valid classifier".format(classifier=classifier))


@cli.command()
@click.option('--port', '-p', default="COM6", help="Serial Port NAME, for example COM3.")
@click.argument('example_name')
@click.argument('args', nargs=-1)
def example(port, example_name, args):
    """
    Run an example from pygarl.examples
    """
    # Load the example
    ex = importlib.import_module("pygarl.examples." + example_name)

    # Check if there are arguments, if so, get them.
    arguments = None
    if len(args) > 0:
        arguments = args[0]

    # Run the example
    ex.run_example(arguments, port=port)


@cli.command()
@click.argument('sample_file')
def plot(sample_file):
    """
    Plot the sample for the given file
    """
    plot_sample(sample_file)

if __name__ == '__main__':
    cli()



from pygarl.base import CallbackManager
from pygarl.classifiers import SVMClassifier
from pygarl.mocks import VerboseMiddleware
from pygarl.data_readers import SerialDataReader
from pygarl.predictors import ClassifierPredictor
from pygarl.sample_managers import DiscreteSampleManager
from string import ascii_lowercase
import pyautogui


def receive_character(character):
    if character != "D":
        pyautogui.typewrite(character)
    else:
        pyautogui.keyDown('backspace')
        pyautogui.keyUp('backspace')


def run_example(*args, **kwargs):
    # Create the SerialDataReader
    sdr = SerialDataReader(kwargs['port'], expected_axis=6, verbose=False)

    # Create the SampleManager
    manager = DiscreteSampleManager()

    # Attach the manager
    sdr.attach_manager(manager)

    # Create a classifier
    classifier = SVMClassifier(model_path=args[0])

    # Load the model
    classifier.load()

    # Print classifier info
    classifier.print_info()

    # Create a ClassifierPredictor
    predictor = ClassifierPredictor(classifier)

    # Attach the classifier predictor
    manager.attach_receiver(predictor)

    # Create a CallbackManager
    callback_mg = CallbackManager(verbose=True)

    # Attach the callback manager
    predictor.attach_callback_manager(callback_mg)

    # Cycle through all characters to bind them
    for c in ascii_lowercase:
        callback_mg.attach_callback(c, receive_character)

    # Attach also space and delete characters
    callback_mg.attach_callback(" ", receive_character)
    callback_mg.attach_callback("D", receive_character)

    # Open the serial connection
    sdr.open()
    print("Opened!")

    # Start the main loop
    sdr.mainloop()



from pygarl.base import Sample
from pygarl.abstracts import AbstractGesturePredictor
import scipy as sp


class HighestAxisPredictor(AbstractGesturePredictor):
    """
    Return the Axis index with the greatest value.
    If absolute_values=True, the predictor will calculate the absolute 
    values before the comparison.
    """
    def __init__(self, absolute_values=False):
        AbstractGesturePredictor.__init__(self)

        # If true, the predictor will calculate the absolute values ( will transform negative numbers into positives )
        self.absolute_values = absolute_values

    def predict(self, sample):
        sample_frames = sample.data.shape[0]  # Get the number of frames in the sample ( rows )

        final_frame = None  # Initialize the array that will contain the frame

        # Check that the sample is not empty
        if sample.data.size > 0:
            # Check the number of frames
            if sample_frames > 1:
                # If there is more than one frame, calculate the average of all frames ( for each axis )
                final_frame = sp.mean(sample.data, axis=0)  # Averaged for each axis
            elif sample_frames == 1:
                # If there is only one frame, set the final frame as the first one in the sample
                final_frame = sample.data[0]
        else:
            # If the sample is empty, raise an exception
            raise ValueError("Can't process an empty sample, it must contain data.")

        # If absolute_values is true, convert the numbers of the frame into their absolute values
        if self.absolute_values:
            final_frame = sp.absolute(final_frame)  # Convert all the numbers to their absolutes

        max_index = sp.argmax(final_frame)  # Find the index of the maximum element in the frame

        # Return the axis index as a string ( a string must be returned as gesture_id )
        return str(max_index)


class ClassifierPredictor(AbstractGesturePredictor):
    """
    Uses a Classifier to predict at which gesture the sample belongs to.
    """
    # TODO: Tests
    def __init__(self, classifier):
        AbstractGesturePredictor.__init__(self)

        # Set the parameters
        self.classifier = classifier

    def predict(self, sample):
        """
        Predict the received sample using the Classifier
        :param sample: sample to predict
        :return: a string containing the gesture_id
        """
        return self.classifier.predict(sample)



from __future__ import print_function
import serial.tools.list_ports


def list_serial_ports():
    """
    Prints all available serial ports
    """
    # Get the ports
    ports = serial.tools.list_ports.comports()

    print("Available Ports:\n")

    # Print all the info
    for i, port in enumerate(ports):
        # Workaround based on the current pySerial version
        # https://stackoverflow.com/questions/43341926/calling-function-method-throws-attritubeerror-tuple-object-has-no-attribute
        if hasattr(port, "device"):
            print(i, ")", port.device, port.description)
        else:
            print(i, ")", port[0], port[1])


import unittest
import shutil
from pygarl.classifiers import SVMClassifier
from pygarl.mocks import *
from pygarl.base import *


class SVMClassifierTestCase(unittest.TestCase):
    """
    Tests to check SVMClassifier consistency.
    In order to test the classifier, this test will try to implement
    a XOR gate using the classifier.
    """

    def setUp(self):
        # Create a test directory if it doesn't exists
        if not os.path.exists("test_dir_svm_classifier"):
            os.makedirs("test_dir_svm_classifier")

        # Create the samples with the 4 possibilities of the XOR gate
        sample1 = Sample(data=[[1], [1]], gesture_id="0")
        sample2 = Sample(data=[[0], [0]], gesture_id="0")
        sample3 = Sample(data=[[1], [0]], gesture_id="1")
        sample4 = Sample(data=[[0], [1]], gesture_id="1")


        # Save the samples in the test directory
        sample1.save_to_file(os.path.join("test_dir_svm_classifier", "0_1.txt"))
        sample2.save_to_file(os.path.join("test_dir_svm_classifier", "0_2.txt"))
        sample3.save_to_file(os.path.join("test_dir_svm_classifier", "1_1.txt"))
        sample4.save_to_file(os.path.join("test_dir_svm_classifier", "1_2.txt"))

        # Initialize an AbstractClassifier
        self.classifier = SVMClassifier(dataset_path="test_dir_svm_classifier", test_size=0.5)

    def tearDown(self):
        # Destroy the test directory
        shutil.rmtree("test_dir_svm_classifier")

        # Destroy the AbstractClassifier
        self.classifier = None

    def test_load_samples_loaded_correctly(self):
        # Load the samples
        self.classifier.load()

        # Test if samples has been loaded correctly

        # Check there are 2 0s and 2 1s
        self.assertEqual(len(list(filter(lambda x: x == 0, self.classifier.y_data))), 2)
        self.assertEqual(len(list(filter(lambda x: x == 1, self.classifier.y_data))), 2)

        # Convert to list from numpy array
        x_data_as_list = map(lambda x: x.tolist(), self.classifier.x_data)

        self.assertIn([1, 1], x_data_as_list)
        self.assertIn([1, 0], x_data_as_list)
        self.assertIn([0, 1], x_data_as_list)
        self.assertIn([0, 0], x_data_as_list)

        # Test if gestures ids has been loaded correctly
        self.assertIn("0", self.classifier.gestures)
        self.assertIn("1", self.classifier.gestures)

    def test_train_model(self):
        self.classifier.load()

        # Load samples multiple times to have a large dataset
        for n in range(100):
            self.classifier.load_samples_data()

        self.assertGreater(self.classifier.train_model(), 0.9)

    def test_predict(self):
        self.classifier.load()

        # Load samples multiple times to have a large dataset
        for n in range(100):
            self.classifier.load_samples_data()

        self.assertGreater(self.classifier.train_model(), 0.9)

        test_sample1 = Sample(data=[[1], [1]])
        test_sample2 = Sample(data=[[1], [0]])

        self.assertEqual(self.classifier.predict(test_sample1), "0")
        self.assertEqual(self.classifier.predict(test_sample2), "1")

    def test_save_model_before_training_should_fail(self):
        self.assertRaises(ValueError, self.classifier.save_model, "path")

    def test_save_and_load_model(self):
        # Load and train the model
        self.classifier.load()

        # Load samples multiple times to have a large dataset
        for n in range(100):
            self.classifier.load_samples_data()

        self.assertGreater(self.classifier.train_model(), 0.9)

        model_path = os.path.join("test_dir_svm_classifier", "model.svm")

        # Save the model
        self.classifier.save_model(model_path)

        # Create a new classifier and load the saved model
        new_classifier = SVMClassifier(model_path=model_path)

        # Load the model from the saved model
        new_classifier.load()

        # Check if the loaded model works correctly
        test_sample1 = Sample(data=[[1], [1]])
        test_sample2 = Sample(data=[[1], [0]])

        self.assertEqual(new_classifier.predict(test_sample1), "0")
        self.assertEqual(new_classifier.predict(test_sample2), "1")


if __name__ == '__main__':
    unittest.main()



import json
import scipy as sp
import numpy as np
import matplotlib.pyplot as plt
from scipy.interpolate import interp2d, interp1d
from sklearn.preprocessing import scale
import pandas as pd


class Sample(object):
    """
    Contains the data recorded from the sensors.
    Provides methods to analyze, manage and persist Samples.
    """

    def __init__(self, data, gesture_id=None):
        self.data = sp.array(data)  # Convert the data to a Numpy array

        # Check that data is a 2-dimensional array
        if self.data.ndim != 2:
            # If not, raise an exception
            raise ValueError("Data must be a 2-dimensional array")

        self.gesture_id = gesture_id

    def save_to_file(self, file_path):
        """
        Save the sample to a file using the JSON format.
        The absolute filename is specified by the "file_path" parameter.
        """
        # Create a dictionary containing all the important data of the sample.
        # NOTE: the numpy array must be converted to a list to serialize it using JSON.
        output_data = {'gesture_id': self.gesture_id, 'data': self.data.tolist()}

        # Save the sample to a file ( filename specified by the file_path param ).
        with open(file_path, 'w') as output_file:
            json.dump(output_data, output_file)

    @staticmethod
    def load_from_file(file_path):
        """
        Return a Sample object by reading a sample file.
        """
        # Open the file and read the content
        with open(file_path) as input_file:
            input_data = json.load(input_file)

        # Create a Sample object with the read data
        sample = Sample(data=input_data['data'], gesture_id=input_data['gesture_id'])

        # Return the Sample object
        return sample

    def get_linearized(self, one_dimensional=False):
        """
        Linearize the data, combining the axes data.
        Useful to feed the data into a machine learning algorithm.
        
        :param one_dimensional: if True, converts the 2-dim [[]] array to a 1-dim [] array
        :return: the linearized array containing the sample data
        """
        # Reshape the data
        output = self.data.reshape(1, -1)

        # If one_dimensional is True, makes the array truly one-dimensional
        if one_dimensional:
            # The array, after the reshape, is in this form [[1, 2, 3]]
            # We take the first axis, so it becomes [1, 2, 3]
            output = output[0]

        return output

    def scale_frames(self, n_frames=50):
        """
        Scales the sample frames, interpolating the data.
        
        :param n_frames: Final number of frames in the sample.
        """
        # Correct the case with only one data frame
        if self.data.shape[0] <= 1:
            # Copy the only frame two times
            # Basically, [[1, 2, 3]] becomes [[1, 2, 3], [1, 2, 3]]
            self.data = sp.repeat(self.data, 2, axis=0)

        # Check the number of axis
        if self.data.shape[1] > 1:  # More than 1 axis
            # Get the Sample data axis dimensions
            x_size = self.data.shape[0]
            y_size = self.data.shape[1]

            # Create the indexes in the axis
            x = sp.arange(0, x_size)
            y = sp.arange(0, y_size)

            # Create a function that interpolates the data points
            f = interp2d(y, x, self.data)

            # Create a new index of the desired size ( n_frames ).
            x_new = sp.linspace(0, x_size - 1, n_frames)

            # Calculate the new interpolated data and change it.
            self.data = f(y, x_new)
        else:  # The case with only one axis must be handled differently
            # Reshape the data
            reshaped = self.data.reshape(1, -1)[0]

            # Get the Sample data axis dimensions
            x_size = self.data.shape[0]

            # Create the indexes in the axis
            x = sp.arange(0, x_size)

            # Create a function that interpolates the data points
            f = interp1d(x, reshaped, kind="zero")

            # Create a new index of the desired size ( n_frames ).
            x_new = sp.linspace(0, x_size - 1, n_frames)

            # Calculate the new interpolated data and reshape it.
            self.data = f(x_new).reshape(-1, 1)

    def framelen(self):
        """
        :return: the number of frames of the sample 
        """
        return self.data.shape[0]

    def subtract(self, amount=0):
        """
        Subtract the amount from all the values
        :param amount: the amount to subtract
        """
        self.data = self.data - amount

    def rolling_mean(self, window):
        """
        Calculate the rolling mean for the sample data
        :param window: rolling mean window
        :return: 
        """
        self.data = pd.rolling_mean(self.data, window, min_periods=1)

    def normalize_frames(self):
        """
        Normalize each axis of the Sample data
        """
        self.data = scale(self.data)

    def abs(self):
        """
        Make each axis of the sample data a positive number calculating the absolute value.
        """
        # Calculate the absolute value of each axis
        self.data = sp.absolute(self.data)

    def trim(self, threshold=100):
        """
        Trim the extremes of the sample data until they exceed the threshold.
        Useful when using a stream and a bit of cleaning is needed.
        """
        # Get the sample gradient
        gradient = self.gradient()

        # Calculate the average for each axis
        average = sp.average(gradient, axis=1)

        # Get the index of the first element grater than the threshold
        initial = sp.argmax(average > threshold)

        # Get a reversed view of the array
        reverse = average[::-1]

        # Get the index of the first element grater than the threshold, starting from the end
        end = average.size - sp.argmax(reverse > threshold)

        # Trim the data array by keeping only the sector between the two indexes
        self.data = self.data[initial:end:]

    def gradient(self):
        """
        Return a numpy array containing the gradient of the sample data
        """
        # Check the number of axis
        if self.data.shape[1] > 1:  # More than 1 axis
            # Calculate the gradient and extract only the first element
            return sp.gradient(self.data)[0]
        else:  # The case with only one axis must be handled differently
            # Reshape the data
            reshaped = self.data.reshape(1, -1)[0]

            # Calculate the gradient and reshape the result
            return sp.gradient(reshaped).reshape(-1, 1)

    def fft(self, append=True):
        """
        Calculates the FFT of the sample data and replace the original data with it.
        """
        # Calculate the real FFT transform
        fourier = np.fft.rfft(self.data, axis=0)

        if fourier.shape[0] > 10:
            # Delete the first term, it's usually too big and covers the other terms
            fourier = fourier[10:]

        # Calculate the absolute value ( complex number argument )
        absolute = sp.absolute(fourier)

        # If append=True, append the fourier transform to the data, if not replace the data
        if append:
            # Append the fft
            self.data = np.append(self.data, absolute, axis=0)
        else:
            # Replace the data
            self.data = absolute

    def plot(self, block=True):
        """
        Using matplotlib, open a dialog with the plotted Sample data.
        :param block:   if true, the plot will be displayed in a non-blocking way
        """
        # Clear the plot
        plt.clf()

        # Add each axis to the plot
        for axis in range(self.data.shape[1]):
            plt.plot(self.data[:, axis], label="AXIS_{axis}".format(axis=axis))

        # Add the axis labels
        plt.xlabel('time', fontsize=18)
        plt.ylabel('value', fontsize=16)
        plt.legend(loc='best', frameon=False)

        # Check if the plot display should be blocking
        if block:  # Blocking
            plt.show()
        else:  # Non Blocking
            # Draw the figure and pause to enable rendering
            plt.draw()
            plt.pause(.001)

            # Show the plot
            plt.show(block=False)

    def __str__(self):
        # Print the data, one frame per line
        return str(self.data)


class CallbackManager(object):
    """
    Receive a gesture_id and call the corresponding callback.
    A callback can be associated to a gesture_id using the attach_callback method.
    """

    def __init__(self, verbose=False):
        # Callback dictionary that associate gesture_id to callback functions
        self.callbacks = {}
        # Set the verbosity of the callback manager
        self.verbose = verbose

    def attach_callback(self, gesture_id, callback):
        """
        Attach a callback to a gesture_id
        """
        self.callbacks[gesture_id] = callback

    def detach_callback(self, gesture_id):
        """
        Detach the gesture from the CallbackManager
        """
        self.callbacks.pop(gesture_id, None)

    def notify_gesture(self, gesture_id):
        """
        Notify a gesture to all the attached callbacks
        """
        # If a callback is set, call it. If not, call the default callback
        if gesture_id in self.callbacks:
            # Call the attached callback
            self.callbacks[gesture_id](gesture_id)
        else:
            # If not set, call the default callback
            self.default_callback(gesture_id)

    def receive_gesture(self, gesture_id):
        """
        Called by a predictor when a new gesture is available
        """
        # If verbose is set, print a notification when a gesture arrives
        if self.verbose:
            print("Received gesture: " + gesture_id)

        # Notify the gesture to all the attached callbacks
        self.notify_gesture(gesture_id)

    def default_callback(self, gesture_id):
        pass



import webbrowser

from pygarl.base import CallbackManager
from pygarl.classifiers import SVMClassifier, MLPClassifier
from pygarl.middlewares import GradientThresholdMiddleware
from pygarl.mocks import VerboseMiddleware
from pygarl.data_readers import SerialDataReader
from pygarl.predictors import ClassifierPredictor
from pygarl.sample_managers import DiscreteSampleManager, StreamSampleManager
from string import ascii_lowercase
import pyautogui

# This example uses a SerialDataReader to read data from a serial port
# and uses a SVMClassifier to translate the gestures
# into characters realizing a gesture keyboard


def receive_gesture(gesture):
    if gesture == "knock":
        pass
    elif gesture == "doubleknock":
        pass


def run_example(*args, **kwargs):
    # Create the SerialDataReader
    sdr = SerialDataReader(kwargs['port'], expected_axis=6, verbose=False)

    # Create the SampleManager
    manager = StreamSampleManager(step=20, window=20)

    # Attach the manager
    sdr.attach_manager(manager)

    # Create a threshold middleware
    middleware = GradientThresholdMiddleware(verbose=False, threshold=10, sample_group_delay=5, group=True)

    # Attach the middleware
    manager.attach_receiver(middleware)

    # Create a classifier
    classifier = SVMClassifier(model_path=args[0])

    # Load the model
    classifier.load()

    # Print classifier info
    classifier.print_info()

    # Create a ClassifierPredictor
    predictor = ClassifierPredictor(classifier)

    # Attach the classifier predictor
    middleware.attach_receiver(predictor)

    # Create a CallbackManager
    callback_mg = CallbackManager(verbose=True)

    # Attach the callback manager
    predictor.attach_callback_manager(callback_mg)

    # Attach the callbacks
    callback_mg.attach_callback("knock", receive_gesture)
    callback_mg.attach_callback("doubleknock", receive_gesture)

    # Open the serial connection
    sdr.open()
    print("Opened!")

    # Start the main loop
    sdr.mainloop()



import joblib
from sklearn import svm
from sklearn import neural_network
from sklearn.model_selection import GridSearchCV
from sklearn.model_selection import train_test_split
from pygarl.abstracts import AbstractClassifier
from sklearn.metrics import confusion_matrix


class SVMClassifier(AbstractClassifier):
    def __init__(self, params=None, n_jobs=8, test_size=0.35, *args, **kwargs):
        AbstractClassifier.__init__(self, *args, **kwargs)

        # Variables that will hold the training data
        self.x_data = []
        self.y_data = []

        # If no parameters are passed, use the default ones
        if params is None:
            params = {'C': [0.001, 0.01, 0.1], 'kernel': ['linear', 'rbf']}

        # Set the classifier parameters
        self.params = params
        self.n_jobs = n_jobs
        self.test_size = test_size

        # Set the verbosity level based on the value of self.verbose
        verbosity = 0
        if self.verbose:
            verbosity = 10

        # Initialize the model
        self.svc = svm.SVC(probability=True)

        # Initialize the GridSearchCV
        self.clf = GridSearchCV(self.svc, self.params, verbose=verbosity, n_jobs=self.n_jobs)

    def load_sample_data(self, sample):
        """
        Process and load a sample before feeding it to the training phase
        
        :param sample: the loaded Sample
        """
        # Transform the data matrix of the sample in a one-dimensional array
        linearized_sample = sample.get_linearized(one_dimensional=True)

        # Get the internal id of the gesture
        internal_id = self.get_internal_id_from_gesture_id(sample.gesture_id)

        # Add the sample data to the list
        self.x_data.append(linearized_sample)
        self.y_data.append(internal_id)

    def train_model(self):
        """
        Train the model using a Grid Search with cross-validation

        :return: the score of the best combination of parameters
        """

        # Split the dataset into two subset, one used for training and one for testing
        X_train, X_test, Y_train, Y_test = train_test_split(self.x_data, self.y_data,
                                                            test_size=self.test_size, random_state=0)

        # Start the training process
        self.clf.fit(X_train, Y_train)

        # Calculates the score of the best estimator found.
        score = self.clf.score(X_test, Y_test)

        # Set the model as trained
        self.is_trained = True

        # Calculates the confusion matrix
        Y_predicted = self.clf.predict(X_test)
        self.confusion_matrix = confusion_matrix(Y_test, Y_predicted)

        # If verbose is True, print the best model found
        if self.verbose:
            print(self.clf.best_estimator_)
            # Print the confusion matrix
            print("Confusion Matrix:")
            print(self.confusion_matrix)

        return score

    def predict_sample(self, sample):
        """
        Return the predicted gesture_id of the specified sample
        
        :param sample: sample used to predict the gesture 
        :return: a string containing the "gesture_id"
        """
        # Linearize the sample data
        linearized_sample = sample.get_linearized()

        # Predict the gesture id with the trained model
        internal_id = self.clf.predict(linearized_sample)

        # Convert the internal_id to the gesture_id string
        gesture_id = self.gestures[internal_id[0]]

        return gesture_id

    def get_attributes(self):
        """
        Return a dictionary containing the needed attributes to save the classifier
        """
        # Get the saves attributes from the Parent Classifier
        attributes = super(SVMClassifier, self).get_attributes()

        # Add the Specific attributes of the classifier
        attributes.update({'clf': self.clf})

        return attributes

    def load_attributes(self, attributes):
        """
        Load the specified attributes in the classifier.
        :param attributes: a dictionary containing the attributes
        """
        # Load the parent attributes
        super(SVMClassifier, self).load_attributes(attributes)

        # Load specific attributes
        self.clf = attributes['clf']


class MLPClassifier(AbstractClassifier):
    def __init__(self, params=None, n_jobs=8, test_size=0.35, *args, **kwargs):
        AbstractClassifier.__init__(self, *args, **kwargs)

        # Variables that will hold the training data
        self.x_data = []
        self.y_data = []

        # If no parameters are passed, use the default ones
        if params is None:
            params = {'alpha': [1e-1, 1e-2, 1e-3, 1e-4, 1e-5, 1e-6, 1e-7], 'solver': ['lbfgs'],
                      'hidden_layer_sizes': [(5, 2), (10, 5)], 'random_state': [1]}

        # Set the classifier parameters
        self.params = params
        self.n_jobs = n_jobs
        self.test_size = test_size

        # Set the verbosity level based on the value of self.verbose
        verbosity = 0
        if self.verbose:
            verbosity = 10

        # Initialize the model
        self.mlp = neural_network.MLPClassifier()

        # Initialize the GridSearchCV
        self.clf = GridSearchCV(self.mlp, self.params, verbose=verbosity, n_jobs=self.n_jobs)

    def load_sample_data(self, sample):
        """
        Process and load a sample before feeding it to the training phase

        :param sample: the loaded Sample
        """
        # Transform the data matrix of the sample in a one-dimensional array
        linearized_sample = sample.get_linearized(one_dimensional=True)

        # Get the internal id of the gesture
        internal_id = self.get_internal_id_from_gesture_id(sample.gesture_id)

        # Add the sample data to the list
        self.x_data.append(linearized_sample)
        self.y_data.append(internal_id)

    def train_model(self):
        """
        Train the model using a Grid Search with cross-validation

        :return: the score of the best combination of parameters
        """

        # Split the dataset into two subset, one used for training and one for testing
        X_train, X_test, Y_train, Y_test = train_test_split(self.x_data, self.y_data,
                                                            test_size=self.test_size, random_state=0)

        # Start the training process
        self.clf.fit(X_train, Y_train)

        # Calculates the score of the best estimator found.
        score = self.clf.score(X_test, Y_test)

        # Set the model as trained
        self.is_trained = True

        # Calculates the confusion matrix
        Y_predicted = self.clf.predict(X_test)
        self.confusion_matrix = confusion_matrix(Y_test, Y_predicted)

        # If verbose is True, print the best model found
        if self.verbose:
            print(self.clf.best_estimator_)
            # Print the confusion matrix
            print("Confusion Matrix:")
            print(self.confusion_matrix)

        return score

    def predict_sample(self, sample):
        """
        Return the predicted gesture_id of the specified sample

        :param sample: sample used to predict the gesture 
        :return: a string containing the "gesture_id"
        """
        # Linearize the sample data
        linearized_sample = sample.get_linearized()

        # Predict the gesture id with the trained model
        internal_id = self.clf.predict(linearized_sample)

        # Convert the internal_id to the gesture_id string
        gesture_id = self.gestures[internal_id]

        return gesture_id

    def get_attributes(self):
        """
        Return a dictionary containing the needed attributes to save the classifier
        """
        # Get the saves attributes from the Parent Classifier
        attributes = super(MLPClassifier, self).get_attributes()

        # Add the Specific attributes of the classifier
        attributes.update({'clf': self.clf})

        return attributes

    def load_attributes(self, attributes):
        """
        Load the specified attributes in the classifier.
        :param attributes: a dictionary containing the attributes
        """
        # Load the parent attributes
        super(MLPClassifier, self).load_attributes(attributes)

        # Load specific attributes
        self.clf = attributes['clf']



from __future__ import print_function
import os
import uuid
import time
import os.path
from pygarl.abstracts import AbstractGestureRecorder
from pygarl.utils import RandomGestureChooser


class FileGestureRecorder(AbstractGestureRecorder):
    def __init__(self, target_dir, max_tries=5, verbose=False, forced_gesture_id=None):
        AbstractGestureRecorder.__init__(self)

        # Make sure the directory is valid, if not, raise an exception
        if not os.path.isdir(target_dir):
            raise ValueError("The specified target directory is not valid.")

        self.target_dir = target_dir
        self.max_tires = max_tries  # Maximum number of saving tries when a filename conflict occur
        self.verbose = verbose
        self.forced_gesture_id = forced_gesture_id

    def receive_sample(self, sample):
        """
        Receive a sample from a Sender and try to save it using the save_sample method
        """
        self.save_sample(sample)

    def save_sample(self, sample):
        """
        Saves the given sample in the target directory
        """
        # If a forced gesture id is set, override the sample gesture id
        if self.forced_gesture_id is not None:
            # If the forced gesture is a value, set it.
            # If it is a gesture chooser, call it and use the return value.
            if not isinstance(self.forced_gesture_id, RandomGestureChooser):
                sample.gesture_id = self.forced_gesture_id
            else:
                sample.gesture_id = self.forced_gesture_id.get_gesture()

        # Make sure the Sample has a gesture_id because it is needed to generate a filename
        if sample.gesture_id is None:
            raise ValueError("The Sample must have a gesture_id to be saved.")

        current_try_count = 0  # Initial number of saving tries

        sample_has_been_saved = False
        filename = None

        # Loop until a valid filename is found or the limit is reached
        while current_try_count < self.max_tires and not sample_has_been_saved:
            # Generate a 6 chars random string using the uuid library
            random_chars = str(uuid.uuid4())[:6].upper()

            # Get the UNIX timestamp as string ( the conversion to int is needed for rounding )
            timestamp = str(int(time.time()))

            # Generate the filename using the gesture_id, the timestamp and the random chars to minimize
            # the probabilities of a filename conflict
            filename = "{id}_{timestamp}_{random}.txt".format(id=sample.gesture_id, timestamp=timestamp,
                                                              random=random_chars)

            # Generate the complete file path
            file_path = os.path.join(self.target_dir, filename)

            # Check if the file already exists
            if not os.path.exists(file_path):
                # If it doesn't exist the name is valid, so save the sample
                sample.save_to_file(file_path)

                sample_has_been_saved = True

            # Increment the try count
            current_try_count += 1

        # If the sample has not been saved, it means that the tries limit has been exceeded
        if not sample_has_been_saved:
            # The tries limit has been exceeded, raise an exception
            raise RuntimeError("Can't find a valid filename for the Sample, conflict limit exceeded.")

        # If verbosity is set to True, print the saved sample filename
        if self.verbose:
            print("SAMPLE SAVED TO FILE:", filename)

        return filename